import os
import sys
import time
from collections import defaultdict
from datetime import datetime, timezone
from itertools import islice
from typing import Type

import pytest

from labelbox import (
    Classification,
    Client,
    Dataset,
    FactCheckingTool,
    MediaType,
    OntologyBuilder,
    Option,
    PromptIssueTool,
    PromptResponseClassification,
    ResponseOption,
    StepReasoningTool,
    Tool,
)
from labelbox.schema.data_row import DataRowMetadataField
from labelbox.schema.ontology_kind import OntologyKind
from labelbox.schema.user import User
from labelbox.schema.tool_building.types import FeatureSchemaAttribute


@pytest.fixture
def constants():
    SPLIT_SCHEMA_ID = "cko8sbczn0002h2dkdaxb5kal"
    TEST_SPLIT_ID = "cko8scbz70005h2dkastwhgqt"
    TEXT_SCHEMA_ID = "cko8s9r5v0001h2dk9elqdidh"
    CAPTURE_DT_SCHEMA_ID = "cko8sdzv70006h2dk8jg64zvb"
    EXPECTED_METADATA_SCHEMA_IDS = [
        SPLIT_SCHEMA_ID,
        TEST_SPLIT_ID,
        TEXT_SCHEMA_ID,
        CAPTURE_DT_SCHEMA_ID,
    ]
    CUSTOM_TEXT_SCHEMA_NAME = "custom_text"

    return {
        "SPLIT_SCHEMA_ID": SPLIT_SCHEMA_ID,
        "TEST_SPLIT_ID": TEST_SPLIT_ID,
        "TEXT_SCHEMA_ID": TEXT_SCHEMA_ID,
        "CAPTURE_DT_SCHEMA_ID": CAPTURE_DT_SCHEMA_ID,
        "EXPECTED_METADATA_SCHEMA_IDS": EXPECTED_METADATA_SCHEMA_IDS,
        "CUSTOM_TEXT_SCHEMA_NAME": CUSTOM_TEXT_SCHEMA_NAME,
    }


@pytest.fixture
def project_based_user(client, rand_gen):
    email = rand_gen(str)
    # Use old mutation because it doesn't require users to accept email invites
    query_str = """mutation MakeNewUserPyApi {
        addMembersToOrganization(
            data: {
                emails: ["%s@labelbox.com"],
                orgRoleId: "%s",
                projectRoles: []
            }
        ) {
        newUserId
        }
    }
    """ % (email, str(client.get_roles()["NONE"].uid))
    user_id = client.execute(query_str)["addMembersToOrganization"][0][
        "newUserId"
    ]
    assert user_id is not None, "Unable to add user with old mutation"
    user = client._get_single(User, user_id)
    yield user
    client.get_organization().remove_user(user)


@pytest.fixture
def project_pack(client):
    projects = [
        client.create_project(
            name=f"user-proj-{idx}",
            media_type=MediaType.Image,
        )
        for idx in range(2)
    ]
    yield projects
    for proj in projects:
        proj.delete()


@pytest.fixture
def project_with_one_feature_ontology(project, client: Client):
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class").asdict(),
    ]
    empty_ontology = {"tools": tools, "classifications": []}
    ontology = client.create_ontology(
        "empty ontology",
        empty_ontology,
        MediaType.Image,
    )
    project.connect_ontology(ontology)
    yield project


@pytest.fixture
def configured_project(
    project_with_one_feature_ontology, initial_dataset, rand_gen, image_url
):
    dataset = initial_dataset
    data_row_id = dataset.create_data_row(row_data=image_url).uid
    project = project_with_one_feature_ontology

    batch = project.create_batch(
        rand_gen(str),
        [data_row_id],  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = [data_row_id]

    yield project

    batch.delete()


@pytest.fixture
def configured_project_with_complex_ontology(
    client: Client, initial_dataset, rand_gen, image_url, teardown_helpers
):
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    dataset = initial_dataset
    data_row = dataset.create_data_row(row_data=image_url)
    data_row_ids = [data_row.uid]

    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids

    ontology = OntologyBuilder()
    tools = [
        Tool(tool=Tool.Type.BBOX, name="test-bbox-class"),
        Tool(tool=Tool.Type.LINE, name="test-line-class"),
        Tool(tool=Tool.Type.POINT, name="test-point-class"),
        Tool(tool=Tool.Type.POLYGON, name="test-polygon-class"),
    ]

    options = [
        Option(value="first option answer"),
        Option(value="second option answer"),
        Option(value="third option answer"),
    ]

    classifications = [
        Classification(
            class_type=Classification.Type.TEXT, name="test-text-class"
        ),
        Classification(
            class_type=Classification.Type.RADIO,
            name="test-radio-class",
            options=options,
        ),
        Classification(
            class_type=Classification.Type.CHECKLIST,
            name="test-checklist-class",
            options=options,
        ),
    ]

    for t in tools:
        for c in classifications:
            t.add_classification(c)
        ontology.add_tool(t)
    for c in classifications:
        ontology.add_classification(c)

    ontology = client.create_ontology(
        "image ontology",
        ontology.asdict(),
        MediaType.Image,
    )
    project.connect_ontology(ontology)

    yield [project, data_row]
    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def ontology(client):
    ontology_builder = OntologyBuilder(
        tools=[
            Tool(tool=Tool.Type.BBOX, name="Box 1", color="#ff0000"),
            Tool(tool=Tool.Type.BBOX, name="Box 2", color="#ff0000"),
        ],
        classifications=[
            Classification(
                name="Root Class",
                class_type=Classification.Type.RADIO,
                options=[
                    Option(value="1", label="Option 1"),
                    Option(value="2", label="Option 2"),
                ],
            )
        ],
    )
    ontology = client.create_ontology(
        "Integration Test Ontology", ontology_builder.asdict(), MediaType.Image
    )
    yield ontology
    client.delete_unused_ontology(ontology.uid)


@pytest.fixture
def video_data(client, rand_gen, video_data_row, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    data_row = dataset.create_data_row(video_data_row)
    data_row = wait_for_data_row_processing(client, data_row)
    data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


def create_video_data_row(rand_gen):
    return {
        "row_data": "https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4",
        "global_key": f"https://storage.googleapis.com/labelbox-datasets/video-sample-data/sample-video-1.mp4-{rand_gen(str)}",
        "media_type": "VIDEO",
    }


@pytest.fixture
def video_data_100_rows(client, rand_gen, wait_for_data_row_processing):
    dataset = client.create_dataset(name=rand_gen(str))
    data_row_ids = []
    for _ in range(100):
        data_row = dataset.create_data_row(create_video_data_row(rand_gen))
        data_row = wait_for_data_row_processing(client, data_row)
        data_row_ids.append(data_row.uid)
    yield dataset, data_row_ids
    dataset.delete()


@pytest.fixture()
def video_data_row(rand_gen):
    return create_video_data_row(rand_gen)


class ExportV2Helpers:
    @classmethod
    def run_project_export_v2_task(
        cls, project, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {
                "project_details": True,
                "performance_details": False,
                "data_row_details": True,
                "label_details": True,
            }
        )
        while num_retries > 0:
            task = project.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break
        return task.result

    @classmethod
    def run_dataset_export_v2_task(
        cls, dataset, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {"performance_details": False, "label_details": True}
        )
        while num_retries > 0:
            task = dataset.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result

    @classmethod
    def run_catalog_export_v2_task(
        cls, client, num_retries=5, task_name=None, filters={}, params={}
    ):
        task = None
        params = (
            params
            if params
            else {"performance_details": False, "label_details": True}
        )
        catalog = client.get_catalog()
        while num_retries > 0:
            task = catalog.export_v2(
                task_name=task_name, filters=filters, params=params
            )
            task.wait_till_done()
            assert task.status == "COMPLETE"
            assert task.errors is None
            if len(task.result) == 0:
                num_retries -= 1
                time.sleep(5)
            else:
                break

        return task.result


@pytest.fixture
def export_v2_test_helpers() -> Type[ExportV2Helpers]:
    return ExportV2Helpers()


@pytest.fixture
def big_dataset_data_row_ids(big_dataset: Dataset):
    export_task = big_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    yield [dr.json["data_row"]["id"] for dr in stream]


@pytest.fixture(scope="function")
def dataset_with_invalid_data_rows(
    unique_dataset: Dataset, upload_invalid_data_rows_for_dataset
):
    upload_invalid_data_rows_for_dataset(unique_dataset)

    yield unique_dataset


@pytest.fixture
def upload_invalid_data_rows_for_dataset():
    def _upload_invalid_data_rows_for_dataset(dataset: Dataset):
        task = dataset.create_data_rows(
            [
                {
                    "row_data": "gs://invalid-bucket/example.png",  # forbidden
                    "external_id": "image-without-access.jpg",
                },
            ]
            * 2
        )
        task.wait_till_done()

    return _upload_invalid_data_rows_for_dataset


@pytest.fixture
def prompt_response_generation_project_with_new_dataset(
    client: Client, rand_gen, request
):
    """fixture is parametrize and needs project_type in request"""
    media_type = request.param
    prompt_response_project = client.create_prompt_response_generation_project(
        name=f"{media_type.value}-{rand_gen(str)}",
        dataset_name=f"{media_type.value}-{rand_gen(str)}",
        data_row_count=1,
        media_type=media_type,
    )

    yield prompt_response_project

    prompt_response_project.delete()


@pytest.fixture
def prompt_response_generation_project_with_dataset_id(
    client: Client, dataset, rand_gen, request
):
    """fixture is parametrized and needs project_type in request"""
    media_type = request.param
    prompt_response_project = client.create_prompt_response_generation_project(
        name=f"{media_type.value}-{rand_gen(str)}",
        dataset_id=dataset.uid,
        data_row_count=1,
        media_type=media_type,
    )

    yield prompt_response_project

    prompt_response_project.delete()


@pytest.fixture
def response_creation_project(client: Client, rand_gen):
    project_name = f"response-creation-project-{rand_gen(str)}"
    project = client.create_response_creation_project(name=project_name)

    yield project

    project.delete()


@pytest.fixture
def prompt_response_features(rand_gen):
    prompt_text = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.PROMPT,
        name=f"{rand_gen(str)}-prompt text",
    )

    response_radio = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.RESPONSE_RADIO,
        name=f"{rand_gen(str)}-response radio classification",
        options=[
            ResponseOption(value=f"{rand_gen(str)}-first radio option answer"),
            ResponseOption(value=f"{rand_gen(str)}-second radio option answer"),
        ],
    )

    response_checklist = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.RESPONSE_CHECKLIST,
        name=f"{rand_gen(str)}-response checklist classification",
        options=[
            ResponseOption(
                value=f"{rand_gen(str)}-first checklist option answer"
            ),
            ResponseOption(
                value=f"{rand_gen(str)}-second checklist option answer"
            ),
        ],
    )

    response_text_with_char = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.RESPONSE_TEXT,
        name=f"{rand_gen(str)}-response text with character min and max",
        character_min=1,
        character_max=10,
    )

    response_text = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.RESPONSE_TEXT,
        name=f"{rand_gen(str)}-response text",
    )

    nested_response_radio = PromptResponseClassification(
        class_type=PromptResponseClassification.Type.RESPONSE_RADIO,
        name=f"{rand_gen(str)}-nested response radio classification",
        options=[
            ResponseOption(
                f"{rand_gen(str)}-first_radio_answer",
                options=[
                    PromptResponseClassification(
                        class_type=PromptResponseClassification.Type.RESPONSE_RADIO,
                        name=f"{rand_gen(str)}-sub_radio_question",
                        options=[
                            ResponseOption(
                                f"{rand_gen(str)}-first_sub_radio_answer"
                            )
                        ],
                    )
                ],
            )
        ],
    )
    yield {
        "prompts": [prompt_text],
        "responses": [
            response_text,
            response_radio,
            response_checklist,
            response_text_with_char,
            nested_response_radio,
        ],
    }


@pytest.fixture
def prompt_response_ontology(
    client: Client, rand_gen, prompt_response_features, request
):
    """fixture is parametrize and needs project_type in request"""

    project_type = request.param
    if project_type == MediaType.LLMPromptCreation:
        ontology_builder = OntologyBuilder(
            tools=[], classifications=prompt_response_features["prompts"]
        )
    elif project_type == MediaType.LLMPromptResponseCreation:
        ontology_builder = OntologyBuilder(
            tools=[],
            classifications=prompt_response_features["prompts"]
            + prompt_response_features["responses"],
        )
    else:
        ontology_builder = OntologyBuilder(
            tools=[], classifications=prompt_response_features["responses"]
        )

    ontology_name = f"prompt-response-{rand_gen(str)}"

    if project_type in MediaType:
        ontology = client.create_ontology(
            ontology_name, ontology_builder.asdict(), media_type=project_type
        )
    else:
        ontology = client.create_ontology(
            ontology_name,
            ontology_builder.asdict(),
            media_type=MediaType.Text,
            ontology_kind=OntologyKind.ResponseCreation,
        )
    yield ontology

    featureSchemaIds = [
        feature["featureSchemaId"]
        for feature in ontology.normalized["classifications"]
    ]

    try:
        client.delete_unused_ontology(ontology.uid)
        for featureSchemaId in featureSchemaIds:
            client.delete_unused_feature_schema(featureSchemaId)
    except Exception as e:
        print(f"Failed to delete ontology {ontology.uid}: {str(e)}")


@pytest.fixture
def point():
    return Tool(
        tool=Tool.Type.POINT,
        name="name",
        color="#ff0000",
    )


@pytest.fixture
def auto_ocr_text_value_class():
    return Classification(
        class_type=Classification.Type.TEXT,
        name="Auto OCR Text Value",
        instructions="Text value for ocr bboxes",
        scope=Classification.Scope.GLOBAL,
        required=False,
        attributes=[
            FeatureSchemaAttribute(
                attributeName="auto-ocr-text-value", attributeValue="true"
            )
        ],
    )


@pytest.fixture
def auto_ocr_bbox(auto_ocr_text_value_class):
    return Tool(
        tool=Tool.Type.BBOX,
        name="Auto ocr bbox",
        color="ff0000",
        attributes=[
            FeatureSchemaAttribute(
                attributeName="auto-ocr", attributeValue="true"
            )
        ],
        classifications=[auto_ocr_text_value_class],
    )


@pytest.fixture
def requires_connection_classification():
    return Classification(
        name="Requires connection radio",
        instructions="Classification that requires a connection",
        class_type=Classification.Type.RADIO,
        attributes=[
            FeatureSchemaAttribute(
                attributeName="requires-connection", attributeValue="true"
            )
        ],
        options=[Option(value="A"), Option(value="B")],
    )


@pytest.fixture
def requires_connection_classification_feature_schema(
    client, requires_connection_classification
):
    created_feature_schema = client.upsert_feature_schema(
        requires_connection_classification.asdict()
    )
    yield created_feature_schema
    client.delete_unused_feature_schema(
        created_feature_schema.normalized["featureSchemaId"]
    )


@pytest.fixture
def auto_ocr_bbox_feature_schema(client, auto_ocr_bbox):
    created_feature_schema = client.upsert_feature_schema(
        auto_ocr_bbox.asdict()
    )
    yield created_feature_schema
    client.delete_unused_feature_schema(
        created_feature_schema.normalized["featureSchemaId"]
    )


@pytest.fixture
def feature_schema(client, point):
    created_feature_schema = client.upsert_feature_schema(point.asdict())

    yield created_feature_schema
    client.delete_unused_feature_schema(
        created_feature_schema.normalized["featureSchemaId"]
    )


@pytest.fixture
def chat_evaluation_ontology(client, rand_gen):
    ontology_name = f"test-chat-evaluation-ontology-{rand_gen(str)}"

    ontology_builder = OntologyBuilder(
        tools=[
            Tool(
                tool=Tool.Type.MESSAGE_SINGLE_SELECTION,
                name="model output single selection",
            ),
            Tool(
                tool=Tool.Type.MESSAGE_MULTI_SELECTION,
                name="model output multi selection",
            ),
            Tool(
                tool=Tool.Type.MESSAGE_RANKING,
                name="model output multi ranking",
            ),
            StepReasoningTool(name="step reasoning"),
            FactCheckingTool(name="fact checking"),
            PromptIssueTool(name="prompt issue"),
        ],
        classifications=[
            Classification(
                class_type=Classification.Type.TEXT,
                name="global model output text classification",
                scope=Classification.Scope.GLOBAL,
            ),
            Classification(
                class_type=Classification.Type.RADIO,
                name="global model output radio classification",
                scope=Classification.Scope.GLOBAL,
                options=[
                    Option(value="global first option answer"),
                    Option(value="global second option answer"),
                ],
            ),
            Classification(
                class_type=Classification.Type.CHECKLIST,
                name="global model output checklist classification",
                scope=Classification.Scope.GLOBAL,
                options=[
                    Option(value="global first option answer"),
                    Option(value="global second option answer"),
                ],
            ),
            Classification(
                class_type=Classification.Type.TEXT,
                name="index model output text classification",
                scope=Classification.Scope.INDEX,
            ),
            Classification(
                class_type=Classification.Type.RADIO,
                name="index model output radio classification",
                scope=Classification.Scope.INDEX,
                options=[
                    Option(value="index first option answer"),
                    Option(value="index second option answer"),
                ],
            ),
            Classification(
                class_type=Classification.Type.CHECKLIST,
                name="index model output checklist classification",
                scope=Classification.Scope.INDEX,
                options=[
                    Option(value="index first option answer"),
                    Option(value="index second option answer"),
                ],
            ),
        ],
    )
    ontology = client.create_ontology(
        ontology_name,
        ontology_builder.asdict(),
        media_type=MediaType.Conversational,
        ontology_kind=OntologyKind.ModelEvaluation,
    )
    yield ontology

    try:
        client.delete_unused_ontology(ontology.uid)
    except Exception as e:
        print(f"Failed to delete ontology {ontology.uid}: {str(e)}")


@pytest.fixture
def live_chat_evaluation_project(client, rand_gen):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    project = client.create_model_evaluation_project(name=project_name)

    yield project

    project.delete()


@pytest.fixture
def live_chat_evaluation_project_with_batch(
    client,
    rand_gen,
    live_chat_evaluation_project,
    offline_conversational_data_row,
):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    project = client.create_model_evaluation_project(name=project_name)

    project.create_batch(
        rand_gen(str),
        [offline_conversational_data_row.uid],  # sample of data row objects
    )

    yield project

    project.delete()


@pytest.fixture
def offline_chat_evaluation_project(client, rand_gen):
    project_name = f"test-offline-model-evaluation-project-{rand_gen(str)}"
    project = client.create_offline_model_evaluation_project(name=project_name)

    yield project

    project.delete()


@pytest.fixture
def chat_evaluation_project_append_to_dataset(client, dataset, rand_gen):
    project_name = f"test-model-evaluation-project-{rand_gen(str)}"
    dataset_id = dataset.uid
    project = client.create_model_evaluation_project(
        name=project_name, dataset_id=dataset_id, data_row_count=1
    )

    yield project

    project.delete()


@pytest.fixture
def offline_conversational_data_row(initial_dataset):
    convo_v2_row_data = {
        "type": "application/vnd.labelbox.conversational.model-chat-evaluation",
        "version": 2,
        "actors": {
            "clxhs9wk000013b6w7imiz0h8": {
                "role": "human",
                "metadata": {"name": "User"},
            },
            "clxhsc6xb00013b6w1awh579j": {
                "role": "model",
                "metadata": {
                    "modelConfigId": "5a50d319-56bd-405d-87bb-4442daea0d0f"
                },
            },
            "clxhsc6xb00023b6wlp0768zs": {
                "role": "model",
                "metadata": {
                    "modelConfigId": "1cfc833a-2684-47df-95ac-bb7d9f9e3e1f"
                },
            },
        },
        "messages": {
            "clxhs9wk000023b6wrufora3k": {
                "actorId": "clxhs9wk000013b6w7imiz0h8",
                "content": [{"type": "text", "content": "Hello world"}],
                "childMessageIds": ["clxhscb4z00033b6wukpvmuol"],
            },
            "clxhscb4z00033b6wukpvmuol": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [
                    {
                        "type": "text",
                        "content": "Hello to you too! 👋 \n\nIt's great to be your guide in the digital world. What can I help you with today? 😊 \n",
                    }
                ],
                "childMessageIds": ["clxhu2s0900013b6wbv0ndddd"],
            },
            "clxhu2s0900013b6wbv0ndddd": {
                "actorId": "clxhs9wk000013b6w7imiz0h8",
                "content": [
                    {
                        "type": "text",
                        "content": "Lets some some multi-turn happening",
                    }
                ],
                "childMessageIds": [
                    "clxhu4qib00023b6wuep47b1l",
                    "clxhu4qib00033b6wf18az01q",
                ],
            },
            "clxhu4qib00023b6wuep47b1l": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [
                    {
                        "type": "text",
                        "content": "Okay, I'm ready for some multi-turn fun! To make it interesting, how about we try building a story together? \n\n**Here's the beginning:**\n\nThe old, dusty book lay forgotten on the shelf, its leather cover cracked and faded. But as the afternoon sun slanted through the window, a single ray caught a glint of gold on the book's spine. Suddenly...\n\n**Now you tell me what happens!** What does the glint of gold turn out to be? What happens next? 🤔 \n",
                    }
                ],
                "childMessageIds": ["clxhu596m00043b6wvkgahcwz"],
            },
            "clxhu4qib00033b6wf18az01q": {
                "actorId": "clxhsc6xb00023b6wlp0768zs",
                "content": [
                    {
                        "type": "text",
                        "content": "Okay, I'm game! Let's have a multi-turn conversation! But to get us started, give me a topic or a question to get us going! 😄 \n\nWhat's on your mind? \n",
                    }
                ],
                "childMessageIds": ["clxhu596m00043b6wvkgahcwz"],
            },
            "clxhu596m00043b6wvkgahcwz": {
                "actorId": "clxhs9wk000013b6w7imiz0h8",
                "content": [
                    {"type": "text", "content": "Can you generate an image?"}
                ],
                "childMessageIds": [
                    "clxhu5hqk00053b6wk21u6xi7",
                    "clxhu5hqk00063b6whl4gofx8",
                ],
            },
            "clxhu5hqk00053b6wk21u6xi7": {
                "actorId": "clxhsc6xb00013b6w1awh579j",
                "content": [
                    {
                        "type": "text",
                        "content": "I can't generate images directly yet. I'm a text-based AI, so my superpowers are all about words! 🪄 \n\nHowever, I'd love to help you describe an image that fits our story. Tell me: \n\n* **What should the image show?** (For example: The dusty book, the glint of gold, the bookshelf, the room, etc.) \n* **What kind of feeling or atmosphere should the image have?** (Mysterious? Magical? Ominous?) \n\nThe more details you give me, the better I can describe the image! ✨ \n",
                    }
                ],
                "childMessageIds": [],
            },
            "clxhu5hqk00063b6whl4gofx8": {
                "actorId": "clxhsc6xb00023b6wlp0768zs",
                "content": [
                    {
                        "type": "text",
                        "content": "I can't *actually* generate images directly. 😔 I'm primarily a text-based AI. \n\nTo help me understand what you'd like to see, tell me: \n\n* **What should be in the image?** Be specific! (e.g., \"A cat wearing a tiny hat\", \"A futuristic cityscape at sunset\")\n* **What style do you imagine?** (e.g., realistic, cartoonish, abstract)\n\nOnce you give me those details, I can try to give you a vivid description that's almost as good as seeing it! 😊 \n",
                    }
                ],
                "childMessageIds": [],
            },
        },
        "rootMessageIds": ["clxhs9wk000023b6wrufora3k"],
    }

    convo_v2_asset = {
        "row_data": convo_v2_row_data,
    }
    data_row = initial_dataset.create_data_row(convo_v2_asset)

    return data_row


@pytest.fixture
def response_data_row(initial_dataset):
    text_asset = {"row_data": "response sample text"}
    data_row = initial_dataset.create_data_row(text_asset)

    return data_row


@pytest.fixture()
def conversation_data_row(initial_dataset, rand_gen):
    data = {
        "row_data": "https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json",
        "global_key": f"https://storage.googleapis.com/labelbox-developer-testing-assets/conversational_text/1000-conversations/conversation-1.json-{rand_gen(str)}",
    }
    convo_asset = {"row_data": data}
    data_row = initial_dataset.create_data_row(convo_asset)

    return data_row


def pytest_configure():
    pytest.report = defaultdict(int)


@pytest.hookimpl(hookwrapper=True)
def pytest_fixture_setup(fixturedef):
    start = time.time()
    yield
    end = time.time()

    exec_time = end - start
    if "FIXTURE_PROFILE" in os.environ:
        pytest.report[fixturedef.argname] += exec_time


@pytest.fixture(scope="session", autouse=True)
def print_perf_summary():
    yield

    if "FIXTURE_PROFILE" in os.environ:
        sorted_dict = dict(
            sorted(
                pytest.report.items(), key=lambda item: item[1], reverse=True
            )
        )
        num_of_entries = 10 if len(sorted_dict) >= 10 else len(sorted_dict)
        slowest_fixtures = [
            (aaa, sorted_dict[aaa])
            for aaa in islice(sorted_dict, num_of_entries)
        ]
        print("\nTop slowest fixtures:\n", slowest_fixtures, file=sys.stderr)


@pytest.fixture
def make_metadata_fields(constants):
    msg = "A message"
    time = datetime.now(timezone.utc)

    fields = [
        DataRowMetadataField(
            schema_id=constants["SPLIT_SCHEMA_ID"],
            value=constants["TEST_SPLIT_ID"],
        ),
        DataRowMetadataField(
            schema_id=constants["CAPTURE_DT_SCHEMA_ID"], value=time
        ),
        DataRowMetadataField(schema_id=constants["TEXT_SCHEMA_ID"], value=msg),
    ]
    return fields



import pytest

from labelbox.data.annotation_types.metrics import (
    ConfusionMatrixAggregation,
    ScalarMetricAggregation,
)
from labelbox.data.annotation_types.metrics import (
    ConfusionMatrixMetric,
    ScalarMetric,
)
from labelbox.data.annotation_types import (
    ScalarMetric,
    Label,
    GenericDataRowData,
)
from labelbox.data.annotation_types.metrics.scalar import RESERVED_METRIC_NAMES
from pydantic import ValidationError


def test_legacy_scalar_metric():
    value = 10
    metric = ScalarMetric(value=value)
    assert metric.value == value

    label = Label(
        data=GenericDataRowData(uid="ckrmd9q8g000009mg6vej7hzg"),
        annotations=[metric],
    )
    expected = {
        "data": {
            "uid": "ckrmd9q8g000009mg6vej7hzg",
        },
        "annotations": [
            {
                "aggregation": ScalarMetricAggregation.ARITHMETIC_MEAN,
                "value": 10.0,
                "extra": {},
            }
        ],
        "extra": {},
        "is_benchmark_reference": False,
    }
    assert label.model_dump(exclude_none=True) == expected


# TODO: Test with confidence


@pytest.mark.parametrize(
    "feature_name,subclass_name,aggregation,value",
    [
        ("cat", "orange", ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
        ("cat", None, ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
        (None, None, ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
        (None, None, None, 0.5),
        ("cat", "orange", ScalarMetricAggregation.ARITHMETIC_MEAN, 0.5),
        ("cat", None, ScalarMetricAggregation.HARMONIC_MEAN, 0.5),
        (None, None, ScalarMetricAggregation.GEOMETRIC_MEAN, 0.5),
        (None, None, ScalarMetricAggregation.SUM, 0.5),
        (
            "cat",
            "orange",
            ScalarMetricAggregation.ARITHMETIC_MEAN,
            {0.1: 0.2, 0.3: 0.5, 0.4: 0.8},
        ),
    ],
)
def test_custom_scalar_metric(feature_name, subclass_name, aggregation, value):
    kwargs = {"aggregation": aggregation} if aggregation is not None else {}
    metric = ScalarMetric(
        metric_name="custom_iou",
        value=value,
        feature_name=feature_name,
        subclass_name=subclass_name,
        **kwargs,
    )
    assert metric.value == value

    label = Label(
        data=GenericDataRowData(uid="ckrmd9q8g000009mg6vej7hzg"),
        annotations=[metric],
    )
    expected = {
        "data": {
            "uid": "ckrmd9q8g000009mg6vej7hzg",
        },
        "annotations": [
            {
                "value": value,
                "metric_name": "custom_iou",
                **({"feature_name": feature_name} if feature_name else {}),
                **({"subclass_name": subclass_name} if subclass_name else {}),
                "aggregation": aggregation
                or ScalarMetricAggregation.ARITHMETIC_MEAN,
                "extra": {},
            }
        ],
        "extra": {},
        "is_benchmark_reference": False,
    }
    assert label.model_dump(exclude_none=True) == expected


@pytest.mark.parametrize(
    "feature_name,subclass_name,aggregation,value",
    [
        (
            "cat",
            "orange",
            ConfusionMatrixAggregation.CONFUSION_MATRIX,
            (0, 1, 2, 3),
        ),
        (
            "cat",
            None,
            ConfusionMatrixAggregation.CONFUSION_MATRIX,
            (0, 1, 2, 3),
        ),
        (None, None, ConfusionMatrixAggregation.CONFUSION_MATRIX, (0, 1, 2, 3)),
        (None, None, None, (0, 1, 2, 3)),
        (
            "cat",
            "orange",
            ConfusionMatrixAggregation.CONFUSION_MATRIX,
            {0.1: (0, 1, 2, 3), 0.3: (0, 1, 2, 3), 0.4: (0, 1, 2, 3)},
        ),
    ],
)
def test_custom_confusison_matrix_metric(
    feature_name, subclass_name, aggregation, value
):
    kwargs = {"aggregation": aggregation} if aggregation is not None else {}
    metric = ConfusionMatrixMetric(
        metric_name="confusion_matrix_50_pct_iou",
        value=value,
        feature_name=feature_name,
        subclass_name=subclass_name,
        **kwargs,
    )
    assert metric.value == value

    label = Label(
        data=GenericDataRowData(uid="ckrmd9q8g000009mg6vej7hzg"),
        annotations=[metric],
    )
    expected = {
        "data": {
            "uid": "ckrmd9q8g000009mg6vej7hzg",
        },
        "annotations": [
            {
                "value": value,
                "metric_name": "confusion_matrix_50_pct_iou",
                **({"feature_name": feature_name} if feature_name else {}),
                **({"subclass_name": subclass_name} if subclass_name else {}),
                "aggregation": aggregation
                or ConfusionMatrixAggregation.CONFUSION_MATRIX,
                "extra": {},
            }
        ],
        "extra": {},
        "is_benchmark_reference": False,
    }
    assert label.model_dump(exclude_none=True) == expected


def test_name_exists():
    # Name is only required for ConfusionMatrixMetric for now.
    with pytest.raises(ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(value=[0, 1, 2, 3])


def test_invalid_aggregations():
    with pytest.raises(ValidationError) as exc_info:
        metric = ScalarMetric(
            metric_name="invalid aggregation",
            value=0.1,
            aggregation=ConfusionMatrixAggregation.CONFUSION_MATRIX,
        )
    with pytest.raises(ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(
            metric_name="invalid aggregation",
            value=[0, 1, 2, 3],
            aggregation=ScalarMetricAggregation.SUM,
        )


def test_invalid_number_of_confidence_scores():
    with pytest.raises(ValidationError) as exc_info:
        metric = ScalarMetric(metric_name="too few scores", value={0.1: 0.1})
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(
            metric_name="too few scores", value={0.1: [0, 1, 2, 3]}
        )
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(ValidationError) as exc_info:
        metric = ScalarMetric(
            metric_name="too many scores",
            value={i / 20.0: 0.1 for i in range(20)},
        )
    assert "Number of confidence scores must be greater" in str(exc_info.value)
    with pytest.raises(ValidationError) as exc_info:
        metric = ConfusionMatrixMetric(
            metric_name="too many scores",
            value={i / 20.0: [0, 1, 2, 3] for i in range(20)},
        )
    assert "Number of confidence scores must be greater" in str(exc_info.value)


@pytest.mark.parametrize("metric_name", RESERVED_METRIC_NAMES)
def test_reserved_names(metric_name: str):
    with pytest.raises(ValidationError) as exc_info:
        ScalarMetric(metric_name=metric_name, value=0.5)
    assert "is a reserved metric name" in exc_info.value.errors()[0]["msg"]



from itertools import product

import pytest
from lbox.exceptions import InconsistentOntologyException

from labelbox import Classification, OntologyBuilder, Option, Tool

_SAMPLE_ONTOLOGY = {
    "tools": [
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "poly",
            "color": "#FF0000",
            "tool": "polygon",
            "classifications": [],
            "attributes": None,
        },
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "segment",
            "color": "#FF0000",
            "tool": "superpixel",
            "classifications": [],
            "attributes": None,
        },
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "bbox",
            "color": "#FF0000",
            "tool": "rectangle",
            "attributes": [
                {
                    "attributeName": "auto-ocr",
                    "attributeValue": "true",
                }
            ],
            "classifications": [
                {
                    "schemaNodeId": None,
                    "featureSchemaId": None,
                    "required": True,
                    "instructions": "nested classification",
                    "name": "nested classification",
                    "type": "radio",
                    "uiMode": "searchable",
                    "options": [
                        {
                            "schemaNodeId": None,
                            "featureSchemaId": None,
                            "label": "first",
                            "value": "first",
                            "options": [
                                {
                                    "schemaNodeId": None,
                                    "featureSchemaId": None,
                                    "required": False,
                                    "instructions": "nested nested text",
                                    "name": "nested nested text",
                                    "type": "text",
                                    "options": [],
                                    "attributes": None,
                                }
                            ],
                        },
                        {
                            "schemaNodeId": None,
                            "featureSchemaId": None,
                            "label": "second",
                            "value": "second",
                            "options": [],
                        },
                    ],
                    "attributes": [
                        {
                            "attributeName": "requires-connection",
                            "attributeValue": "true",
                        }
                    ],
                },
                {
                    "schemaNodeId": None,
                    "featureSchemaId": None,
                    "required": True,
                    "instructions": "nested text",
                    "name": "nested text",
                    "type": "text",
                    "options": [],
                    "attributes": None,
                },
            ],
        },
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "dot",
            "color": "#FF0000",
            "tool": "point",
            "classifications": [],
            "attributes": None,
        },
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "polyline",
            "color": "#FF0000",
            "tool": "line",
            "classifications": [],
            "attributes": None,
        },
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": False,
            "name": "ner",
            "color": "#FF0000",
            "tool": "named-entity",
            "classifications": [],
            "attributes": None,
        },
    ],
    "classifications": [
        {
            "schemaNodeId": None,
            "featureSchemaId": None,
            "required": True,
            "instructions": "This is a question.",
            "name": "This is a question.",
            "type": "radio",
            "scope": "global",
            "uiMode": "searchable",
            "attributes": None,
            "options": [
                {
                    "schemaNodeId": None,
                    "featureSchemaId": None,
                    "label": "yes",
                    "value": "definitely yes",
                    "options": [],
                },
                {
                    "schemaNodeId": None,
                    "featureSchemaId": None,
                    "label": "no",
                    "value": "definitely not",
                    "options": [],
                },
            ],
        }
    ],
}


@pytest.mark.parametrize("tool_type", list(Tool.Type))
def test_create_tool(tool_type) -> None:
    t = Tool(tool=tool_type, name="tool")
    assert t.tool == tool_type


@pytest.mark.parametrize("class_type", list(Classification.Type))
def test_create_classification(class_type) -> None:
    c = Classification(class_type=class_type, name="classification")
    assert c.class_type == class_type


@pytest.mark.parametrize(
    "ui_mode_type, class_type",
    list(product(list(Classification.UIMode), list(Classification.Type))),
)
def test_create_classification_with_ui_mode(ui_mode_type, class_type) -> None:
    c = Classification(
        name="classification", class_type=class_type, ui_mode=ui_mode_type
    )
    assert c.ui_mode == ui_mode_type


@pytest.mark.parametrize(
    "value, expected_value, typing", [(3, 3, int), ("string", "string", str)]
)
def test_create_option_with_value(value, expected_value, typing) -> None:
    o = Option(value=value)
    assert o.value == expected_value
    assert o.value == o.label


@pytest.mark.parametrize(
    "value, label, expected_value, typing",
    [(3, 2, 3, int), ("string", "another string", "string", str)],
)
def test_create_option_with_value_and_label(
    value, label, expected_value, typing
) -> None:
    o = Option(value=value, label=label)
    assert o.value == expected_value
    assert o.value != o.label
    assert isinstance(o.value, typing)


def test_create_empty_ontology() -> None:
    o = OntologyBuilder()
    assert o.tools == []
    assert o.classifications == []


def test_add_ontology_tool() -> None:
    o = OntologyBuilder()
    o.add_tool(Tool(tool=Tool.Type.BBOX, name="bounding box"))

    second_tool = Tool(tool=Tool.Type.SEGMENTATION, name="segmentation")
    o.add_tool(second_tool)
    assert len(o.tools) == 2

    for tool in o.tools:
        assert type(tool) is Tool

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_tool(Tool(tool=Tool.Type.BBOX, name="bounding box"))
    assert "Duplicate tool name" in str(exc.value)


def test_add_ontology_classification() -> None:
    o = OntologyBuilder()
    o.add_classification(
        Classification(class_type=Classification.Type.TEXT, name="text")
    )

    second_classification = Classification(
        class_type=Classification.Type.CHECKLIST, name="checklist"
    )
    o.add_classification(second_classification)
    assert len(o.classifications) == 2

    for classification in o.classifications:
        assert type(classification) is Classification

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_classification(
            Classification(class_type=Classification.Type.TEXT, name="text")
        )
    assert "Duplicate classification name" in str(exc.value)


def test_tool_add_classification() -> None:
    t = Tool(tool=Tool.Type.SEGMENTATION, name="segmentation")
    c = Classification(class_type=Classification.Type.TEXT, name="text")
    t.add_classification(c)
    assert t.classifications == [c]

    with pytest.raises(Exception) as exc:
        t.add_classification(c)
    assert "Duplicate nested classification" in str(exc)


def test_classification_add_option() -> None:
    c = Classification(class_type=Classification.Type.RADIO, name="radio")
    o = Option(value="option")
    c.add_option(o)
    assert c.options == [o]

    with pytest.raises(InconsistentOntologyException) as exc:
        c.add_option(Option(value="option"))
    assert "Duplicate option" in str(exc.value)


def test_option_add_option() -> None:
    o = Option(value="option")
    c = Classification(class_type=Classification.Type.TEXT, name="text")
    o.add_option(c)
    assert o.options == [c]

    with pytest.raises(InconsistentOntologyException) as exc:
        o.add_option(c)
    assert "Duplicate nested classification" in str(exc.value)


def test_ontology_asdict() -> None:
    assert (
        OntologyBuilder.from_dict(_SAMPLE_ONTOLOGY).asdict() == _SAMPLE_ONTOLOGY
    )


def test_classification_using_instructions_instead_of_name_shows_warning():
    with pytest.warns(Warning):
        Classification(class_type=Classification.Type.TEXT, instructions="text")


def test_classification_without_name_raises_error():
    with pytest.raises(ValueError):
        Classification(class_type=Classification.Type.TEXT)



import pytest
from lbox.exceptions import LabelboxError, OperationNotAllowedException


def test_live_chat_evaluation_project(
    live_chat_evaluation_project, model_config
):
    project = live_chat_evaluation_project

    project.set_project_model_setup_complete()
    assert bool(project.model_setup_complete) is True

    with pytest.raises(
        expected_exception=LabelboxError,
        match="Cannot create model config for project because model setup is complete",
    ):
        project.add_model_config(model_config.uid)


def test_live_chat_evaluation_project_delete_cofig(
    live_chat_evaluation_project, model_config
):
    project = live_chat_evaluation_project
    project_model_config_id = project.add_model_config(model_config.uid)
    assert project_model_config_id

    project_model_config = None
    for pmg in project.project_model_configs():
        if pmg.uid == project_model_config_id:
            project_model_config = pmg
            break
    assert project_model_config

    project.set_project_model_setup_complete()
    assert bool(project.model_setup_complete) is True

    with pytest.raises(
        expected_exception=LabelboxError,
        match="Cannot create model config for project because model setup is complete",
    ):
        project_model_config.delete()


def test_offline_chat_evaluation_project(
    offline_chat_evaluation_project, model_config
):
    project = offline_chat_evaluation_project

    with pytest.raises(
        expected_exception=OperationNotAllowedException,
        match="Only live model chat evaluation projects can complete model setup",
    ):
        project.set_project_model_setup_complete()


def test_any_other_project(project, model_config):
    with pytest.raises(
        expected_exception=OperationNotAllowedException,
        match="Only live model chat evaluation projects can complete model setup",
    ):
        project.set_project_model_setup_complete()



import time
import uuid

import pytest

from labelbox import Client, MediaType
from labelbox.schema.annotation_import import AnnotationImportState, LabelImport
from labelbox.schema.media_type import MediaType


@pytest.fixture
def ontology(client: Client):
    bbox_tool_with_nested_text = {
        "required": False,
        "name": "bbox_tool_with_nested_text",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [
            {
                "required": False,
                "instructions": "nested",
                "name": "nested",
                "type": "radio",
                "options": [
                    {
                        "label": "radio_option_1",
                        "value": "radio_value_1",
                        "options": [
                            {
                                "required": False,
                                "instructions": "nested_checkbox",
                                "name": "nested_checkbox",
                                "type": "checklist",
                                "options": [
                                    {
                                        "label": "nested_checkbox_option_1",
                                        "value": "nested_checkbox_value_1",
                                        "options": [],
                                    },
                                    {
                                        "label": "nested_checkbox_option_2",
                                        "value": "nested_checkbox_value_2",
                                    },
                                ],
                            },
                            {
                                "required": False,
                                "instructions": "nested_text",
                                "name": "nested_text",
                                "type": "text",
                                "options": [],
                            },
                        ],
                    },
                ],
            }
        ],
    }

    bbox_tool = {
        "required": False,
        "name": "bbox",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [
            {
                "required": False,
                "instructions": "nested",
                "name": "nested",
                "type": "radio",
                "options": [
                    {
                        "label": "radio_option_1",
                        "value": "radio_value_1",
                        "options": [
                            {
                                "required": False,
                                "instructions": "nested_checkbox",
                                "name": "nested_checkbox",
                                "type": "checklist",
                                "options": [
                                    {
                                        "label": "nested_checkbox_option_1",
                                        "value": "nested_checkbox_value_1",
                                        "options": [],
                                    },
                                    {
                                        "label": "nested_checkbox_option_2",
                                        "value": "nested_checkbox_value_2",
                                    },
                                ],
                            }
                        ],
                    },
                ],
            }
        ],
    }

    polygon_tool = {
        "required": False,
        "name": "polygon",
        "tool": "polygon",
        "color": "#FF34FF",
        "classifications": [],
    }
    polyline_tool = {
        "required": False,
        "name": "polyline",
        "tool": "line",
        "color": "#FF4A46",
        "classifications": [],
    }
    point_tool = {
        "required": False,
        "name": "point--",
        "tool": "point",
        "color": "#008941",
        "classifications": [],
    }
    raster_segmentation_tool = {
        "required": False,
        "name": "segmentation_mask",
        "tool": "raster-segmentation",
        "color": "#ff0000",
        "classifications": [],
    }
    checklist = {
        "required": False,
        "instructions": "checklist",
        "name": "checklist",
        "type": "checklist",
        "options": [
            {"label": "option1", "value": "option1"},
            {"label": "option2", "value": "option2"},
            {"label": "optionN", "value": "optionn"},
        ],
    }

    free_form_text = {
        "required": False,
        "instructions": "text",
        "name": "text",
        "type": "text",
        "options": [],
    }

    radio = {
        "required": False,
        "instructions": "radio",
        "name": "radio",
        "type": "radio",
        "options": [
            {
                "label": "first_radio_answer",
                "value": "first_radio_answer",
                "options": [],
            },
            {
                "label": "second_radio_answer",
                "value": "second_radio_answer",
                "options": [],
            },
        ],
    }

    tools = [
        bbox_tool,
        bbox_tool_with_nested_text,
        polygon_tool,
        polyline_tool,
        point_tool,
        raster_segmentation_tool,
    ]
    classifications = [
        checklist,
        free_form_text,
        radio,
    ]
    ontology = client.create_ontology(
        "image ontology",
        {"tools": tools, "classifications": classifications},
        MediaType.Image,
    )
    return ontology


@pytest.fixture
def video_ontology(client: Client):
    bbox_tool_with_nested_text = {
        "required": False,
        "name": "bbox_tool_with_nested_text",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [
            {
                "required": False,
                "instructions": "nested",
                "name": "nested",
                "type": "radio",
                "options": [
                    {
                        "label": "radio_option_1",
                        "value": "radio_value_1",
                        "options": [
                            {
                                "required": False,
                                "instructions": "nested_checkbox",
                                "name": "nested_checkbox",
                                "type": "checklist",
                                "options": [
                                    {
                                        "label": "nested_checkbox_option_1",
                                        "value": "nested_checkbox_value_1",
                                        "options": [],
                                    },
                                    {
                                        "label": "nested_checkbox_option_2",
                                        "value": "nested_checkbox_value_2",
                                    },
                                ],
                            },
                            {
                                "required": False,
                                "instructions": "nested_text",
                                "name": "nested_text",
                                "type": "text",
                                "options": [],
                            },
                        ],
                    },
                ],
            }
        ],
    }

    bbox_tool = {
        "required": False,
        "name": "bbox",
        "tool": "rectangle",
        "color": "#a23030",
        "classifications": [
            {
                "required": False,
                "instructions": "nested",
                "name": "nested",
                "type": "radio",
                "options": [
                    {
                        "label": "radio_option_1",
                        "value": "radio_value_1",
                        "options": [
                            {
                                "required": False,
                                "instructions": "nested_checkbox",
                                "name": "nested_checkbox",
                                "type": "checklist",
                                "options": [
                                    {
                                        "label": "nested_checkbox_option_1",
                                        "value": "nested_checkbox_value_1",
                                        "options": [],
                                    },
                                    {
                                        "label": "nested_checkbox_option_2",
                                        "value": "nested_checkbox_value_2",
                                    },
                                ],
                            }
                        ],
                    },
                ],
            }
        ],
    }

    polyline_tool = {
        "required": False,
        "name": "polyline",
        "tool": "line",
        "color": "#FF4A46",
        "classifications": [],
    }
    point_tool = {
        "required": False,
        "name": "point--",
        "tool": "point",
        "color": "#008941",
        "classifications": [],
    }
    raster_segmentation_tool = {
        "required": False,
        "name": "segmentation_mask",
        "tool": "raster-segmentation",
        "color": "#ff0000",
        "classifications": [],
    }
    checklist = {
        "required": False,
        "instructions": "checklist",
        "name": "checklist",
        "type": "checklist",
        "options": [
            {"label": "option1", "value": "option1"},
            {"label": "option2", "value": "option2"},
            {"label": "optionN", "value": "optionn"},
        ],
    }
    checklist_index = {
        "required": False,
        "instructions": "checklist_index",
        "name": "checklist_index",
        "type": "checklist",
        "scope": "index",
        "options": [
            {"label": "option1_index", "value": "option1_index"},
            {"label": "option2_index", "value": "option2_index"},
            {"label": "optionN_index", "value": "optionn_index"},
        ],
    }

    free_form_text = {
        "required": False,
        "instructions": "text",
        "name": "text",
        "type": "text",
        "options": [],
    }

    radio = {
        "required": False,
        "instructions": "radio",
        "name": "radio",
        "type": "radio",
        "options": [
            {
                "label": "first_radio_answer",
                "value": "first_radio_answer",
                "options": [],
            },
            {
                "label": "second_radio_answer",
                "value": "second_radio_answer",
                "options": [],
            },
        ],
    }

    tools = [
        bbox_tool,
        bbox_tool_with_nested_text,
        polyline_tool,
        point_tool,
        raster_segmentation_tool,
    ]
    classifications = [
        checklist_index,
        checklist,
        free_form_text,
        radio,
    ]
    ontology = client.create_ontology(
        "image ontology",
        {"tools": tools, "classifications": classifications},
        MediaType.Video,
    )
    return ontology


@pytest.fixture
def polygon_inference(prediction_id_mapping):
    polygon = prediction_id_mapping["polygon"].copy()
    polygon.update(
        {
            "polygon": [
                {"x": 147.692, "y": 118.154},
                {"x": 142.769, "y": 104.923},
                {"x": 57.846, "y": 118.769},
                {"x": 28.308, "y": 169.846},
            ]
        }
    )
    del polygon["tool"]
    return polygon


@pytest.fixture
def configured_project_with_ontology(
    client, initial_dataset, ontology, rand_gen, image_url, teardown_helpers
):
    dataset = initial_dataset
    project = client.create_project(
        name=rand_gen(str),
        media_type=MediaType.Image,
    )
    project.connect_ontology(ontology)
    data_row_ids = []

    normalized_ontology = ontology.normalized

    for _ in range(
        len(normalized_ontology["tools"])
        + len(normalized_ontology["classifications"])
    ):
        data_row_ids.append(dataset.create_data_row(row_data=image_url).uid)
    project.create_batch(
        rand_gen(str),
        data_row_ids,  # sample of data row objects
        5,  # priority between 1(Highest) - 5(lowest)
    )
    project.data_row_ids = data_row_ids
    yield project
    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def configured_project_without_data_rows(
    client, ontology, rand_gen, teardown_helpers
):
    project = client.create_project(
        name=rand_gen(str),
        description=rand_gen(str),
        media_type=MediaType.Image,
    )

    project.connect_ontology(ontology)
    yield project
    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def configured_video_project_without_data_rows(
    client, video_ontology, rand_gen, teardown_helpers
):
    project = client.create_project(
        name=rand_gen(str),
        description=rand_gen(str),
        media_type=MediaType.Video,
    )

    project.connect_ontology(video_ontology)
    yield project
    teardown_helpers.teardown_project_labels_ontology_feature_schemas(project)


@pytest.fixture
def model_run_with_data_rows(
    client,
    configured_project_with_ontology,
    model_run_predictions,
    model_run,
    wait_for_label_processing,
):
    configured_project_with_ontology.enable_model_assisted_labeling()
    use_data_row_ids = [p["dataRow"]["id"] for p in model_run_predictions]
    model_run.upsert_data_rows(use_data_row_ids)

    upload_task = LabelImport.create_from_objects(
        client,
        configured_project_with_ontology.uid,
        f"label-import-{uuid.uuid4()}",
        model_run_predictions,
    )
    upload_task.wait_until_done()
    assert (
        upload_task.state == AnnotationImportState.FINISHED
    ), "Label Import did not finish"
    assert (
        len(upload_task.errors) == 0
    ), f"Label Import {upload_task.name} failed with errors {upload_task.errors}"
    labels = wait_for_label_processing(configured_project_with_ontology)
    label_ids = [label.uid for label in labels]
    model_run.upsert_labels(label_ids)
    yield model_run, labels
    model_run.delete()
    # TODO: Delete resources when that is possible ..


@pytest.fixture
def model_run_predictions(
    polygon_inference, rectangle_inference, line_inference
):
    # Not supporting mask since there isn't a signed url representing a seg mask to upload
    return [polygon_inference, rectangle_inference, line_inference]


@pytest.fixture
def model(client, rand_gen, configured_project):
    ontology = configured_project.ontology()
    data = {"name": rand_gen(str), "ontology_id": ontology.uid}
    model = client.create_model(data["name"], data["ontology_id"])
    yield model
    try:
        model.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def model_run(rand_gen, model):
    name = rand_gen(str)
    model_run = model.create_model_run(name)
    yield model_run
    try:
        model_run.delete()
    except:
        # Already was deleted by the test
        pass


@pytest.fixture
def wait_for_label_processing():
    """
    Do not use. Only for testing.

    Returns project's labels as a list after waiting for them to finish processing.
    If `project.labels()` is called before label is fully processed,
    it may return an empty set
    """

    def func(project):
        timeout_seconds = 10
        while True:
            labels = list(project.labels())
            if len(labels) > 0:
                return labels
            timeout_seconds -= 2
            if timeout_seconds <= 0:
                raise TimeoutError(
                    f"Timed out waiting for label for project '{project.uid}' to finish processing"
                )
            time.sleep(2)

    return func


@pytest.fixture
def prediction_id_mapping(configured_project_with_ontology):
    # Maps tool types to feature schema ids
    project = configured_project_with_ontology
    ontology = project.ontology().normalized
    result = {}

    for idx, tool in enumerate(ontology["tools"] + ontology["classifications"]):
        if "tool" in tool:
            tool_type = tool["tool"]
        else:
            tool_type = (
                tool["type"]
                if "scope" not in tool
                else f"{tool['type']}_{tool['scope']}"
            )  # so 'checklist' of 'checklist_index'

        # TODO: remove this once we have a better way to associate multiple tools instances with a single tool type
        if tool_type == "rectangle":
            value = {
                "uuid": str(uuid.uuid4()),
                "schemaId": tool["featureSchemaId"],
                "name": tool["name"],
                "dataRow": {
                    "id": project.data_row_ids[idx],
                },
                "tool": tool,
            }
            if tool_type not in result:
                result[tool_type] = []
            result[tool_type].append(value)
        else:
            result[tool_type] = {
                "uuid": str(uuid.uuid4()),
                "schemaId": tool["featureSchemaId"],
                "name": tool["name"],
                "dataRow": {
                    "id": project.data_row_ids[idx],
                },
                "tool": tool,
            }
    return result


@pytest.fixture
def line_inference(prediction_id_mapping):
    line = prediction_id_mapping["line"].copy()
    line.update(
        {"line": [{"x": 147.692, "y": 118.154}, {"x": 150.692, "y": 160.154}]}
    )
    del line["tool"]
    return line


@pytest.fixture
def polygon_inference(prediction_id_mapping):
    polygon = prediction_id_mapping["polygon"].copy()
    polygon.update(
        {
            "polygon": [
                {"x": 147.692, "y": 118.154},
                {"x": 142.769, "y": 104.923},
                {"x": 57.846, "y": 118.769},
                {"x": 28.308, "y": 169.846},
            ]
        }
    )
    del polygon["tool"]
    return polygon


def find_tool_by_name(tool_instances, name):
    for tool in tool_instances:
        if tool["name"] == name:
            return tool
    return None


@pytest.fixture
def rectangle_inference(prediction_id_mapping):
    tool_instance = find_tool_by_name(
        prediction_id_mapping["rectangle"], "bbox"
    )
    rectangle = tool_instance.copy()
    rectangle.update(
        {
            "bbox": {"top": 48, "left": 58, "height": 65, "width": 12},
            "classifications": [
                {
                    "schemaId": rectangle["tool"]["classifications"][0][
                        "featureSchemaId"
                    ],
                    "name": rectangle["tool"]["classifications"][0]["name"],
                    "answer": {
                        "schemaId": rectangle["tool"]["classifications"][0][
                            "options"
                        ][0]["featureSchemaId"],
                        "name": rectangle["tool"]["classifications"][0][
                            "options"
                        ][0]["value"],
                    },
                }
            ],
        }
    )
    del rectangle["tool"]
    return rectangle



import json
import logging
import time
import warnings
from collections import namedtuple
from datetime import datetime, timezone
from string import Template
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    get_args,
)

from labelbox.schema.annotation_import import LabelImport, MALPredictionImport
from lbox.exceptions import (
    InvalidQueryError,
    LabelboxError,
    ProcessingWaitTimeout,
    ResourceNotFoundError,
    error_message_for_unparsed_graphql_error,
)  # type: ignore

from labelbox import utils
from labelbox.orm import query
from labelbox.orm.db_object import DbObject, Deletable, Updateable
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.pagination import PaginatedCollection
from labelbox.schema.consensus_settings import ConsensusSettings
from labelbox.schema.create_batches_task import CreateBatchesTask
from labelbox.schema.data_row import DataRow
from labelbox.schema.export_filters import (
    ProjectExportFilters,
    build_filters,
)
from labelbox.schema.export_params import ProjectExportParams
from labelbox.schema.export_task import ExportTask
from labelbox.schema.identifiable import DataRowIdentifier
from labelbox.schema.identifiables import (
    DataRowIdentifiers,
)
from labelbox.schema.labeling_service import (
    LabelingService,
    LabelingServiceStatus,
)
from labelbox.schema.labeling_service_dashboard import LabelingServiceDashboard
from labelbox.schema.media_type import MediaType
from labelbox.schema.model_config import ModelConfig
from labelbox.schema.ontology_kind import (
    EditorTaskType,
    UploadType,
)
from labelbox.schema.project_model_config import ProjectModelConfig
from labelbox.schema.project_overview import (
    ProjectOverview,
    ProjectOverviewDetailed,
)
from labelbox.schema.resource_tag import ResourceTag
from labelbox.schema.task import Task
from labelbox.schema.task_queue import TaskQueue

if TYPE_CHECKING:
    pass


DataRowPriority = int
LabelingParameterOverrideInput = Tuple[DataRowIdentifier, DataRowPriority]

logger = logging.getLogger(__name__)
MAX_SYNC_BATCH_ROW_COUNT = 1_000


def validate_labeling_parameter_overrides(
    data: List[LabelingParameterOverrideInput],
) -> None:
    for idx, row in enumerate(data):
        data_row_identifier = row[0]
        priority = row[1]
        if not isinstance(data_row_identifier, get_args(DataRowIdentifier)):
            raise TypeError(
                f"Data row identifier should be of type DataRowIdentifier. Found {type(data_row_identifier)}."
            )
        if len(row) < 2:
            raise TypeError(
                f"Data must be a list of tuples each containing two elements: a  DataRowIdentifier and priority (int). Found {len(row)} items. Index: {idx}"
            )
        if not isinstance(priority, int):
            id = data_row_identifier.key
            raise TypeError(
                f"Priority must be an int. Found {type(priority)} for data_row_identifier {id}"
            )


class Project(DbObject, Updateable, Deletable):
    """A Project is a container that includes a labeling frontend, an ontology,
    datasets and labels.

    Attributes:
        name (str)
        description (str)
        updated_at (datetime)
        created_at (datetime)
        setup_complete (datetime)
        last_activity_time (datetime)
        auto_audit_number_of_labels (int)
        auto_audit_percentage (float)
        is_benchmark_enabled (bool)
        is_consensus_enabled (bool)

        created_by (Relationship): `ToOne` relationship to User
        organization (Relationship): `ToOne` relationship to Organization
        labeling_frontend (Relationship): `ToOne` relationship to LabelingFrontend
        labeling_frontend_options (Relationship): `ToMany` relationship to LabelingFrontendOptions
        labeling_parameter_overrides (Relationship): `ToMany` relationship to LabelingParameterOverride
        webhooks (Relationship): `ToMany` relationship to Webhook
        benchmarks (Relationship): `ToMany` relationship to Benchmark
        ontology (Relationship): `ToOne` relationship to Ontology
        task_queues (Relationship): `ToMany` relationship to TaskQueue
    """

    name = Field.String("name")
    description = Field.String("description")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    setup_complete = Field.DateTime("setup_complete")
    last_activity_time = Field.DateTime("last_activity_time")
    auto_audit_number_of_labels = Field.Int("auto_audit_number_of_labels")
    auto_audit_percentage = Field.Float("auto_audit_percentage")
    # Bind data_type and allowedMediaTYpe using the GraphQL type MediaType
    media_type = Field.Enum(MediaType, "media_type", "allowedMediaType")
    editor_task_type = Field.Enum(EditorTaskType, "editor_task_type")
    data_row_count = Field.Int("data_row_count")
    model_setup_complete: Field = Field.Boolean("model_setup_complete")
    upload_type: Field = Field.Enum(UploadType, "upload_type")
    is_benchmark_enabled = Field.Boolean("is_benchmark_enabled")
    is_consensus_enabled = Field.Boolean("is_consensus_enabled")

    # Relationships
    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization", False)
    labeling_frontend = Relationship.ToOne(
        "LabelingFrontend",
        config=Relationship.Config(disconnect_supported=False),
    )
    labeling_frontend_options = Relationship.ToMany(
        "LabelingFrontendOptions", False, "labeling_frontend_options"
    )
    labeling_parameter_overrides = Relationship.ToMany(
        "LabelingParameterOverride", False, "labeling_parameter_overrides"
    )
    webhooks = Relationship.ToMany("Webhook", False)
    benchmarks = Relationship.ToMany("Benchmark", False)
    ontology = Relationship.ToOne("Ontology", True)

    #
    _wait_processing_max_seconds = 3600

    def is_chat_evaluation(self) -> bool:
        """
        Returns:
            True if this project is a live chat evaluation project, False otherwise
        """
        return (
            self.media_type == MediaType.Conversational
            and self.editor_task_type == EditorTaskType.ModelChatEvaluation
        )

    def is_prompt_response(self) -> bool:
        """
        Returns:
            True if this project is a prompt response project, False otherwise
        """
        return (
            self.media_type == MediaType.LLMPromptResponseCreation
            or self.media_type == MediaType.LLMPromptCreation
            or self.editor_task_type == EditorTaskType.ResponseCreation
        )

    def is_auto_data_generation(self) -> bool:
        return self.upload_type == UploadType.Auto  # type: ignore

    # we test not only the project ontology is None, but also a default empty ontology that we create when we attach a labeling front end in createLabelingFrontendOptions
    def is_empty_ontology(self) -> bool:
        ontology = self.ontology()  # type: ignore
        return ontology is None or (
            len(ontology.tools()) == 0 and len(ontology.classifications()) == 0
        )

    def project_model_configs(self):
        query_str = """query ProjectModelConfigsPyApi($id: ID!) {
            project(where: {id : $id}) {
                projectModelConfigs {
                    id
                    modelConfigId
                    modelConfig {
                        id
                        modelId
                        inferenceParams
                    }
                    projectId
                }
            }
        }"""
        data = {"id": self.uid}
        res = self.client.execute(query_str, data)
        return [
            ProjectModelConfig(self.client, projectModelConfig)
            for projectModelConfig in res["project"]["projectModelConfigs"]
        ]

    def update(self, **kwargs):
        """Updates this project with the specified attributes

        Args:
            kwargs: a dictionary containing attributes to be upserted

        Note that the queue_mode cannot be changed after a project has been created.

        Additionally, the quality setting cannot be changed after a project has been created. The quality mode
        for a project is inferred through the following attributes:

        Benchmark:
            auto_audit_number_of_labels = 1 and auto_audit_percentage = 1.0

        Consensus:
            auto_audit_number_of_labels > 1 or auto_audit_percentage <= 1.0

        Attempting to switch between benchmark and consensus modes is an invalid operation and will result
        in an error.
        """

        media_type = kwargs.get("media_type")
        if media_type:
            if MediaType.is_supported(media_type):
                kwargs["media_type"] = media_type.value
            else:
                raise TypeError(
                    f"{media_type} is not a valid media type. Use"
                    f" any of {MediaType.get_supported_members()}"
                    " from MediaType. Example: MediaType.Image."
                )

        return super().update(**kwargs)

    def members(self) -> PaginatedCollection:
        """Fetch all current members for this project

        Returns:
            A `PaginatedCollection` of `ProjectMember`s

        """
        id_param = "projectId"
        query_str = """query ProjectMemberOverviewPyApi($%s: ID!) {
             project(where: {id : $%s}) { id members(skip: %%d first: %%d){ id user { %s } role { id name } accessFrom }
           }
        }""" % (id_param, id_param, query.results_query_part(Entity.User))
        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: str(self.uid)},
            ["project", "members"],
            ProjectMember,
        )

    def update_project_resource_tags(
        self, resource_tag_ids: List[str]
    ) -> List[ResourceTag]:
        """Creates project resource tags

        Args:
            resource_tag_ids
        Returns:
            a list of ResourceTag ids that was created.
        """
        project_id_param = "projectId"
        tag_ids_param = "resourceTagIds"

        query_str = """mutation UpdateProjectResourceTagsPyApi($%s:ID!,$%s:[String!]) {
            project(where:{id:$%s}){updateProjectResourceTags(input:{%s:$%s}){%s}}}""" % (
            project_id_param,
            tag_ids_param,
            project_id_param,
            tag_ids_param,
            tag_ids_param,
            query.results_query_part(ResourceTag),
        )

        res = self.client.execute(
            query_str,
            {project_id_param: self.uid, tag_ids_param: resource_tag_ids},
        )

        return [
            ResourceTag(self.client, tag)
            for tag in res["project"]["updateProjectResourceTags"]
        ]

    def get_resource_tags(self) -> List[ResourceTag]:
        """
        Returns tags for a project
        """
        query_str = """query GetProjectResourceTagsPyApi($projectId: ID!) {
            project(where: {id: $projectId}) {
                name
                resourceTags {%s}
            }
            }""" % (query.results_query_part(ResourceTag))

        results = self.client.execute(query_str, {"projectId": self.uid})[
            "project"
        ]["resourceTags"]

        return [ResourceTag(self.client, tag) for tag in results]

    def labels(self, datasets=None, order_by=None) -> PaginatedCollection:
        """Custom relationship expansion method to support limited filtering.

        Args:
            datasets (iterable of Dataset): Optional collection of Datasets
                whose Labels are sought. If not provided, all Labels in
                this Project are returned.
            order_by (None or (Field, Field.Order)): Ordering clause.
        """
        Label = Entity.Label

        if datasets is not None:
            where = " where:{dataRow: {dataset: {id_in: [%s]}}}" % ", ".join(
                '"%s"' % dataset.uid for dataset in datasets
            )
        else:
            where = ""

        if order_by is not None:
            query.check_order_by_clause(Label, order_by)
            order_by_str = "orderBy: %s_%s" % (
                order_by[0].graphql_name,
                order_by[1].name.upper(),
            )
        else:
            order_by_str = ""

        id_param = "projectId"
        query_str = """query GetProjectLabelsPyApi($%s: ID!)
            {project (where: {id: $%s})
                {labels (skip: %%d first: %%d %s %s) {%s}}}""" % (
            id_param,
            id_param,
            where,
            order_by_str,
            query.results_query_part(Label),
        )

        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: self.uid},
            ["project", "labels"],
            Label,
        )

    def export(
        self,
        task_name: Optional[str] = None,
        filters: Optional[ProjectExportFilters] = None,
        params: Optional[ProjectExportParams] = None,
    ) -> ExportTask:
        """
        Creates a project export task with the given params and returns the task.

        >>>     task = project.export(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "data_row_ids": [DATA_ROW_ID_1, DATA_ROW_ID_2, ...] # or global_keys: [DATA_ROW_GLOBAL_KEY_1, DATA_ROW_GLOBAL_KEY_2, ...]
        >>>             "batch_ids": [BATCH_ID_1, BATCH_ID_2, ...]
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """
        task, _ = self._export(task_name, filters, params, streamable=True)
        return ExportTask(task)

    def export_v2(
        self,
        task_name: Optional[str] = None,
        filters: Optional[ProjectExportFilters] = None,
        params: Optional[ProjectExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a project export task with the given params and returns the task.

        For more information visit: https://docs.labelbox.com/docs/exports-v2#export-from-a-project-python-sdk

        >>>     task = project.export_v2(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "data_row_ids": [DATA_ROW_ID_1, DATA_ROW_ID_2, ...] # or global_keys: [DATA_ROW_GLOBAL_KEY_1, DATA_ROW_GLOBAL_KEY_2, ...]
        >>>             "batch_ids": [BATCH_ID_1, BATCH_ID_2, ...]
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """

        warnings.warn(
            "The method export_v2 for Project is deprecated and will be removed in the next major release. Use the export method instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        task, is_streamable = self._export(task_name, filters, params)
        if is_streamable:
            return ExportTask(task, True)
        return task

    def _export(
        self,
        task_name: Optional[str] = None,
        filters: Optional[ProjectExportFilters] = None,
        params: Optional[ProjectExportParams] = None,
        streamable: bool = False,
    ) -> Tuple[Task, bool]:
        _params = params or ProjectExportParams(
            {
                "attachments": False,
                "embeddings": False,
                "metadata_fields": False,
                "data_row_details": False,
                "project_details": False,
                "performance_details": False,
                "label_details": False,
                "media_type_override": None,
                "interpolated_frames": False,
            }
        )

        _filters = filters or ProjectExportFilters(
            {
                "last_activity_at": None,
                "label_created_at": None,
                "data_row_ids": None,
                "global_keys": None,
                "batch_ids": None,
                "workflow_status": None,
            }
        )

        mutation_name = "exportDataRowsInProject"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInProjectInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}"
        )

        media_type_override = _params.get("media_type_override", None)
        query_params: Dict[str, Any] = {
            "input": {
                "taskName": task_name,
                "isStreamableReady": True,
                "filters": {
                    "projectId": self.uid,
                    "searchQuery": {
                        "scope": None,
                        "query": [],
                    },
                },
                "params": {
                    "mediaTypeOverride": media_type_override.value
                    if media_type_override is not None
                    else None,
                    "includeAttachments": _params.get("attachments", False),
                    "includeEmbeddings": _params.get("embeddings", False),
                    "includeMetadata": _params.get("metadata_fields", False),
                    "includeDataRowDetails": _params.get(
                        "data_row_details", False
                    ),
                    "includeProjectDetails": _params.get(
                        "project_details", False
                    ),
                    "includePerformanceDetails": _params.get(
                        "performance_details", False
                    ),
                    "includeLabelDetails": _params.get("label_details", False),
                    "includeInterpolatedFrames": _params.get(
                        "interpolated_frames", False
                    ),
                },
                "streamable": streamable,
            }
        }

        search_query = build_filters(self.client, _filters)
        query_params["input"]["filters"]["searchQuery"]["query"] = search_query

        res = self.client.execute(
            create_task_query_str, query_params, error_log_key="errors"
        )
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(self.client, task_id), is_streamable

    def export_issues(self, status=None) -> str:
        """Calls the server-side Issues exporting that
        returns the URL to that payload.

        Args:
            status (string): valid values: Open, Resolved
        Returns:
            URL of the data file with this Project's issues.
        """
        id_param = "projectId"
        status_param = "status"
        query_str = """query GetProjectIssuesExportPyApi($%s: ID!, $%s: IssueStatus) {
            project(where: { id: $%s }) {
                issueExportUrl(where: { status: $%s })
            }
        }""" % (id_param, status_param, id_param, status_param)

        valid_statuses = {None, "Open", "Resolved"}

        if status not in valid_statuses:
            raise ValueError(
                "status must be in {}. Found {}".format(valid_statuses, status)
            )

        res = self.client.execute(
            query_str, {id_param: self.uid, status_param: status}
        )

        res = res["project"]

        logger.debug("Project '%s' issues export, link generated", self.uid)

        return res.get("issueExportUrl")

    def upsert_instructions(self, instructions_file: str) -> None:
        """
        * Uploads instructions to the UI. Running more than once will replace the instructions

        Args:
            instructions_file (str): Path to a local file.
                * Must be a pdf or html file

        Raises:
            ValueError:
                * project must be setup
                * instructions file must have a ".pdf" or ".html" extension
        """

        if self.setup_complete is None:
            raise ValueError(
                "Cannot attach instructions to a project that has not been set up."
            )

        frontend = self.labeling_frontend()

        if frontend.name != "Editor":
            logger.warning(
                "This function has only been tested to work with the Editor front end. Found %s",
                frontend.name,
            )

        supported_instruction_formats = (".pdf", ".html")
        if not instructions_file.endswith(supported_instruction_formats):
            raise ValueError(
                f"instructions_file must be a pdf or html file. Found {instructions_file}"
            )

        instructions_url = self.client.upload_file(instructions_file)

        query_str = """mutation setprojectinsructionsPyApi($projectId: ID!, $instructions_url: String!) {
                setProjectInstructions(
                    where: {id: $projectId},
                    data: {instructionsUrl: $instructions_url}
                ) {
                    id
                    ontology {
                    id
                    options
                    }
                }
            }"""

        self.client.execute(
            query_str,
            {"projectId": self.uid, "instructions_url": instructions_url},
        )

    def labeler_performance(self) -> PaginatedCollection:
        """Returns the labeler performances for this Project.

        Returns:
            A PaginatedCollection of LabelerPerformance objects.
        """
        id_param = "projectId"
        query_str = """query LabelerPerformancePyApi($%s: ID!) {
            project(where: {id: $%s}) {
                labelerPerformance(skip: %%d first: %%d) {
                    count user {%s} secondsPerLabel totalTimeLabeling consensus
                    averageBenchmarkAgreement lastActivityTime}
            }}""" % (id_param, id_param, query.results_query_part(Entity.User))

        def create_labeler_performance(client, result):
            result["user"] = Entity.User(client, result["user"])
            # python isoformat doesn't accept Z as utc timezone
            result["lastActivityTime"] = utils.format_iso_from_string(
                result["lastActivityTime"].replace("Z", "+00:00")
            )
            return LabelerPerformance(
                **{
                    utils.snake_case(key): value
                    for key, value in result.items()
                }
            )

        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: self.uid},
            ["project", "labelerPerformance"],
            create_labeler_performance,
        )

    def review_metrics(self, net_score) -> int:
        """Returns this Project's review metrics.

        Args:
            net_score (None or Review.NetScore): Indicates desired metric.
        Returns:
            int, aggregation count of reviews for given `net_score`.
        """
        if net_score not in (None,) + tuple(Entity.Review.NetScore):
            raise InvalidQueryError(
                "Review metrics net score must be either None "
                "or one of Review.NetScore values"
            )
        id_param = "projectId"
        net_score_literal = "None" if net_score is None else net_score.name
        query_str = """query ProjectReviewMetricsPyApi($%s: ID!){
            project(where: {id:$%s})
            {reviewMetrics {labelAggregate(netScore: %s) {count}}}
        }""" % (id_param, id_param, net_score_literal)
        res = self.client.execute(query_str, {id_param: self.uid})
        return res["project"]["reviewMetrics"]["labelAggregate"]["count"]

    def connect_ontology(self, ontology) -> None:
        """
        Connects the ontology to the project. If an editor is not setup, it will be connected as well.

        Note: For live chat model evaluation projects, the editor setup is skipped because it is automatically setup when the project is created.

        Args:
            ontology (Ontology): The ontology to attach to the project
        """
        if not self.is_empty_ontology():
            raise ValueError("Ontology already connected to project.")

        if (
            self.labeling_frontend() is None
        ):  # Chat evaluation projects are automatically set up via the same api that creates a project
            self._connect_default_labeling_front_end(
                ontology_as_dict={"tools": [], "classifications": []}
            )

        query_str = """mutation ConnectOntologyPyApi($projectId: ID!, $ontologyId: ID!){
            project(where: {id: $projectId}) {connectOntology(ontologyId: $ontologyId) {id}}}"""
        self.client.execute(
            query_str, {"ontologyId": ontology.uid, "projectId": self.uid}
        )
        timestamp = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        self.update(setup_complete=timestamp)

    def _connect_default_labeling_front_end(self, ontology_as_dict: dict):
        labeling_frontend = self.labeling_frontend()
        if (
            labeling_frontend is None
        ):  # Chat evaluation projects are automatically set up via the same api that creates a project
            warnings.warn("Connecting default labeling editor for the project.")
            labeling_frontend = next(
                self.client._get_labeling_frontends(
                    where=Entity.LabelingFrontend.name == "Editor"
                )
            )
            self.labeling_frontend.connect(labeling_frontend)

        if not isinstance(ontology_as_dict, str):
            labeling_frontend_options_str = json.dumps(ontology_as_dict)
        else:
            labeling_frontend_options_str = ontology_as_dict

        LFO = Entity.LabelingFrontendOptions
        self.client._create(
            LFO,
            {
                LFO.project: self,
                LFO.labeling_frontend: labeling_frontend,
                LFO.customization_options: labeling_frontend_options_str,
            },
        )

    def get_mal_prediction_imports(self) -> PaginatedCollection:
        """Returns mal prediction import objects which are used in model-assisted labeling associated with the project.

        Returns:
            PaginatedCollection
        """

        id_param = "projectId"
        query_str = """
        query getModelAssistedLabelingPredictionImportsPyApi($%s: ID!) {
            modelAssistedLabelingPredictionImports(skip: %%d, first: %%d, where: { projectId: $%s  }) { %s }}
        """ % (
            id_param,
            id_param,
            query.results_query_part(MALPredictionImport),
        )

        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: self.uid},
            ["modelAssistedLabelingPredictionImports"],
            MALPredictionImport,
        )

    def get_label_imports(self) -> PaginatedCollection:
        """Returns label import objects associated with the project.

        Returns:
            PaginatedCollection
        """

        id_param = "projectId"
        query_str = """
        query getLabelImportsPyApi($%s: ID!) {
            labelImports(skip: %%d, first: %%d, where: { projectId: $%s  }) { %s }}
        """ % (
            id_param,
            id_param,
            query.results_query_part(LabelImport),
        )

        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: self.uid},
            ["labelImports"],
            LabelImport,
        )

    def create_batch(
        self,
        name: str,
        data_rows: Optional[List[Union[str, DataRow]]] = None,
        priority: int = 5,
        consensus_settings: Optional[Dict[str, Any]] = None,
        global_keys: Optional[List[str]] = None,
    ):
        """
        Creates a new batch for a project. One of `global_keys` or `data_rows` must be provided, but not both. A
            maximum of 100,000 data rows can be added to a batch.

        Args:
            name: a name for the batch, must be unique within a project
            data_rows: Either a list of `DataRows` or Data Row ids.
            global_keys: global keys for data rows to add to the batch.
            priority: An optional priority for the Data Rows in the Batch. 1 highest -> 5 lowest
            consensus_settings: An optional dictionary with consensus settings: {'number_of_labels': 3,
                'coverage_percentage': 0.1}

        Returns: the created batch

        Raises:
            lbox.exceptions.ValueError if a project is not batch mode, if the project is auto data generation, if the batch exceeds 100k data rows
        """

        if (
            self.is_auto_data_generation() and not self.is_chat_evaluation()
        ):  # NOTE live chat evaluatiuon projects in sdk do not pre-generate data rows, but use batch as all other projects
            raise ValueError(
                "Cannot create batches for auto data generation projects"
            )

        dr_ids = []
        if data_rows is not None:
            for dr in data_rows:
                if isinstance(dr, Entity.DataRow):
                    dr_ids.append(dr.uid)
                elif isinstance(dr, str):
                    dr_ids.append(dr)
                else:
                    raise ValueError(
                        "`data_rows` must be DataRow ids or DataRow objects"
                    )

        if data_rows is not None:
            row_count = len(dr_ids)
        elif global_keys is not None:
            row_count = len(global_keys)
        else:
            row_count = 0

        if row_count > 100_000:
            raise ValueError(
                "Batch exceeds max size, break into smaller batches"
            )
        if not row_count:
            raise ValueError("You need at least one data row in a batch")

        self._wait_until_data_rows_are_processed(
            dr_ids, global_keys, self._wait_processing_max_seconds
        )

        if consensus_settings:
            consensus_settings = ConsensusSettings(
                **consensus_settings
            ).model_dump(by_alias=True)

        if row_count >= MAX_SYNC_BATCH_ROW_COUNT:
            return self._create_batch_async(
                name, dr_ids, global_keys, priority, consensus_settings
            )
        else:
            return self._create_batch_sync(
                name, dr_ids, global_keys, priority, consensus_settings
            )

    def create_batches(
        self,
        name_prefix: str,
        data_rows: Optional[List[Union[str, DataRow]]] = None,
        global_keys: Optional[List[str]] = None,
        priority: int = 5,
        consensus_settings: Optional[Dict[str, Any]] = None,
    ) -> CreateBatchesTask:
        """
        Creates batches for a project from a list of data rows. One of `global_keys` or `data_rows` must be provided,
        but not both. When more than 100k data rows are specified and thus multiple batches are needed, the specific
        batch that each data row will be placed in is undefined.

        Batches will be created with the specified name prefix and a unique suffix. The suffix will be a 4-digit
        number starting at 0000. For example, if the name prefix is "batch" and 3 batches are created, the names
        will be "batch0000", "batch0001", and "batch0002". This method will throw an error if a batch with the same
        name already exists.

        Args:
            name_prefix: a prefix for the batch names, must be unique within a project
            data_rows: Either a list of `DataRows` or Data Row ids.
            global_keys: global keys for data rows to add to the batch.
            priority: An optional priority for the Data Rows in the Batch. 1 highest -> 5 lowest
            consensus_settings: An optional dictionary with consensus settings: {'number_of_labels': 3,
                'coverage_percentage': 0.1}

        Returns: a task for the created batches
        """

        dr_ids = []
        if data_rows is not None:
            for dr in data_rows:
                if isinstance(dr, Entity.DataRow):
                    dr_ids.append(dr.uid)
                elif isinstance(dr, str):
                    dr_ids.append(dr)
                else:
                    raise ValueError(
                        "`data_rows` must be DataRow ids or DataRow objects"
                    )

        self._wait_until_data_rows_are_processed(
            dr_ids, global_keys, self._wait_processing_max_seconds
        )

        if consensus_settings:
            consensus_settings = ConsensusSettings(
                **consensus_settings
            ).model_dump(by_alias=True)

        method = "createBatches"
        mutation_str = """mutation %sPyApi($projectId: ID!, $input: CreateBatchesInput!) {
                                      project(where: {id: $projectId}) {
                                        %s(input: $input) {
                                          tasks {
                                            batchUuid
                                            taskId
                                          }
                                        }
                                      }
                                    }
                                """ % (method, method)

        params = {
            "projectId": self.uid,
            "input": {
                "batchNamePrefix": name_prefix,
                "dataRowIds": dr_ids,
                "globalKeys": global_keys,
                "priority": priority,
                "consensusSettings": consensus_settings,
            },
        }

        tasks = self.client.execute(mutation_str, params)["project"][method][
            "tasks"
        ]
        batch_ids = [task["batchUuid"] for task in tasks]
        task_ids = [task["taskId"] for task in tasks]

        return CreateBatchesTask(self.client, self.uid, batch_ids, task_ids)

    def create_batches_from_dataset(
        self,
        name_prefix: str,
        dataset_id: str,
        priority: int = 5,
        consensus_settings: Optional[Dict[str, Any]] = None,
    ) -> CreateBatchesTask:
        """
        Creates batches for a project from a dataset, selecting only the data rows that are not already added to the
        project. When the dataset contains more than 100k data rows and multiple batches are needed, the specific batch
        that each data row will be placed in is undefined. Note that data rows may not be immediately available for a
        project after being added to a dataset; use the `_wait_until_data_rows_are_processed` method to ensure that
        data rows are available before creating batches.

        Batches will be created with the specified name prefix and a unique suffix. The suffix will be a 4-digit
        number starting at 0000. For example, if the name prefix is "batch" and 3 batches are created, the names
        will be "batch0000", "batch0001", and "batch0002". This method will throw an error if a batch with the same
        name already exists.

        Args:
            name_prefix: a prefix for the batch names, must be unique within a project
            dataset_id: the id of the dataset to create batches from
            priority: An optional priority for the Data Rows in the Batch. 1 highest -> 5 lowest
            consensus_settings: An optional dictionary with consensus settings: {'number_of_labels': 3,
                'coverage_percentage': 0.1}

        Returns: a task for the created batches
        """

        if consensus_settings:
            consensus_settings = ConsensusSettings(
                **consensus_settings
            ).model_dump(by_alias=True)

        method = "createBatchesFromDataset"
        mutation_str = """mutation %sPyApi($projectId: ID!, $input: CreateBatchesFromDatasetInput!) {
                                        project(where: {id: $projectId}) {
                                            %s(input: $input) {
                                              tasks {
                                                batchUuid
                                                taskId
                                              }
                                            }
                                        }
                                    }
                                """ % (method, method)

        params = {
            "projectId": self.uid,
            "input": {
                "batchNamePrefix": name_prefix,
                "datasetId": dataset_id,
                "priority": priority,
                "consensusSettings": consensus_settings,
            },
        }

        tasks = self.client.execute(mutation_str, params)["project"][method][
            "tasks"
        ]

        batch_ids = [task["batchUuid"] for task in tasks]
        task_ids = [task["taskId"] for task in tasks]

        return CreateBatchesTask(self.client, self.uid, batch_ids, task_ids)

    def _create_batch_sync(
        self, name, dr_ids, global_keys, priority, consensus_settings
    ):
        method = "createBatchV2"
        query_str = """mutation %sPyApi($projectId: ID!, $batchInput: CreateBatchInput!) {
                  project(where: {id: $projectId}) {
                    %s(input: $batchInput) {
                        batch {
                            %s
                        }
                        failedDataRowIds
                    }
                  }
                }
            """ % (method, method, query.results_query_part(Entity.Batch))
        params = {
            "projectId": self.uid,
            "batchInput": {
                "name": name,
                "dataRowIds": dr_ids,
                "globalKeys": global_keys,
                "priority": priority,
                "consensusSettings": consensus_settings,
            },
        }
        res = self.client.execute(
            query_str, params, timeout=180.0, experimental=True
        )["project"][method]
        batch = res["batch"]
        batch["size"] = res["batch"]["size"]
        return Entity.Batch(
            self.client,
            self.uid,
            batch,
            failed_data_row_ids=res["failedDataRowIds"],
        )

    def _create_batch_async(
        self,
        name: str,
        dr_ids: Optional[List[str]] = None,
        global_keys: Optional[List[str]] = None,
        priority: int = 5,
        consensus_settings: Optional[Dict[str, float]] = None,
    ):
        method = "createEmptyBatch"
        create_empty_batch_mutation_str = """mutation %sPyApi($projectId: ID!, $input: CreateEmptyBatchInput!) {
                                      project(where: {id: $projectId}) {
                                        %s(input: $input) {
                                            id
                                        }
                                      }
                                    }
                                """ % (method, method)

        params = {
            "projectId": self.uid,
            "input": {"name": name, "consensusSettings": consensus_settings},
        }

        res = self.client.execute(
            create_empty_batch_mutation_str,
            params,
            timeout=180.0,
            experimental=True,
        )["project"][method]
        batch_id = res["id"]

        method = "addDataRowsToBatchAsync"
        add_data_rows_mutation_str = """mutation %sPyApi($projectId: ID!, $input: AddDataRowsToBatchInput!) {
                                      project(where: {id: $projectId}) {
                                        %s(input: $input) {
                                          taskId
                                        }
                                      }
                                    }
                                """ % (method, method)

        params = {
            "projectId": self.uid,
            "input": {
                "batchId": batch_id,
                "dataRowIds": dr_ids,
                "globalKeys": global_keys,
                "priority": priority,
            },
        }

        res = self.client.execute(
            add_data_rows_mutation_str, params, timeout=180.0, experimental=True
        )["project"][method]

        task_id = res["taskId"]

        task = self._wait_for_task(task_id)
        if task.status != "COMPLETE":
            raise LabelboxError(
                "Batch was not created successfully: " + json.dumps(task.errors)
            )

        return self.client.get_batch(self.uid, batch_id)

    def get_label_count(self) -> int:
        """
        Returns: the total number of labels in this project.
        """

        query_str = """query LabelCountPyApi($projectId: ID!) {
            project(where: {id: $projectId}) {
                labelCount
            }
        }"""

        res = self.client.execute(query_str, {"projectId": self.uid})
        return res["project"]["labelCount"]

    def add_model_config(self, model_config_id: str) -> str:
        """Adds a model config to this project.

        Args:
            model_config_id (str): ID of a model config to add to this project.

        Returns:
            str, ID of the project model config association. This is needed for updating and deleting associations.
        """

        query = """mutation CreateProjectModelConfigPyApi($projectId: ID!, $modelConfigId: ID!)  {
                    createProjectModelConfig(input: {projectId: $projectId, modelConfigId: $modelConfigId}) {
                        projectModelConfigId
                    }
                }"""

        params = {
            "projectId": self.uid,
            "modelConfigId": model_config_id,
        }
        try:
            result = self.client.execute(query, params)
        except LabelboxError as e:
            if e.message.startswith(
                "Unknown error: "
            ):  # unfortunate hack to handle unparsed graphql errors
                error_content = error_message_for_unparsed_graphql_error(
                    e.message
                )
            else:
                error_content = e.message
            raise LabelboxError(message=error_content) from e

        if not result:
            raise ResourceNotFoundError(ModelConfig, params)
        return result["createProjectModelConfig"]["projectModelConfigId"]

    def delete_project_model_config(self, project_model_config_id: str) -> bool:
        """Deletes the association between a model config and this project.

        Args:
            project_model_config_id (str): ID of a project model config association to delete for this project.

        Returns:
            bool, indicates if the operation was a success.
        """
        query = """mutation DeleteProjectModelConfigPyApi($id: ID!)  {
                    deleteProjectModelConfig(input: {id: $id}) {
                        success
                    }
                }"""

        params = {
            "id": project_model_config_id,
        }
        result = self.client.execute(query, params)
        if not result:
            raise ResourceNotFoundError(ProjectModelConfig, params)
        return result["deleteProjectModelConfig"]["success"]

    def set_project_model_setup_complete(self) -> bool:
        """
        Sets the model setup is complete for this project.
        Once the project is marked as "setup complete", a user can not add  / modify delete existing project model configs.

        Returns:
            bool, indicates if the model setup is complete.

        NOTE: This method should only be used for live model evaluation projects.
            It will throw exception for all other types of projects.
            User Project is_chat_evaluation() method to check if the project is a live model evaluation project.
        """
        query = """mutation SetProjectModelSetupCompletePyApi($projectId: ID!) {
            setProjectModelSetupComplete(where: {id: $projectId}, data: {modelSetupComplete: true}) {
                modelSetupComplete
            }
        }"""

        result = self.client.execute(query, {"projectId": self.uid})
        self.model_setup_complete = result["setProjectModelSetupComplete"][
            "modelSetupComplete"
        ]
        return result["setProjectModelSetupComplete"]["modelSetupComplete"]

    def set_labeling_parameter_overrides(
        self, data: List[LabelingParameterOverrideInput]
    ) -> bool:
        """Adds labeling parameter overrides to this project.

        See information on priority here:
            https://docs.labelbox.com/en/configure-editor/queue-system#reservation-system

            >>> project.set_labeling_parameter_overrides([
            >>>     (data_row_gk1, 2), (data_row_gk2, 1)])

        Args:
            data (iterable): An iterable of tuples. Each tuple must contain
                (DataRowIdentifier, priority<int>) for the new override.
                DataRowIdentifier is an object representing a data row id or a global key. A DataIdentifier object can be a UniqueIds or GlobalKeys class.

                Priority:
                    * Data will be labeled in priority order.
                        - A lower number priority is labeled first.
                        - All signed 32-bit integers are accepted, from -2147483648 to 2147483647.
                    * Priority is not the queue position.
                        - The position is determined by the relative priority.
                        - E.g. [(data_row_1, 5,1), (data_row_2, 2,1), (data_row_3, 10,1)]
                            will be assigned in the following order: [data_row_2, data_row_1, data_row_3]
                    * The priority only effects items in the queue.
                        - Assigning a priority will not automatically add the item back into the queue.
        Returns:
            bool, indicates if the operation was a success.
        """
        data = [t[:2] for t in data]
        validate_labeling_parameter_overrides(data)

        template = Template(
            """mutation SetLabelingParameterOverridesPyApi($$projectId: ID!)
                {project(where: { id: $$projectId })
                {setLabelingParameterOverrides
                (dataWithDataRowIdentifiers: [$dataWithDataRowIdentifiers])
                {success}}}
            """
        )

        data_rows_with_identifiers = ""
        for data_row, priority in data:
            data_rows_with_identifiers += f'{{dataRowIdentifier: {{id: "{data_row.key}", idType: {data_row.id_type}}}, priority: {priority}}},'

        query_str = template.substitute(
            dataWithDataRowIdentifiers=data_rows_with_identifiers
        )
        res = self.client.execute(query_str, {"projectId": self.uid})
        return res["project"]["setLabelingParameterOverrides"]["success"]

    def update_data_row_labeling_priority(
        self,
        data_rows: DataRowIdentifiers,
        priority: int,
    ) -> bool:
        """
        Updates labeling parameter overrides to this project in bulk. This method allows up to 1 million data rows to be
        updated at once.

        See information on priority here:
            https://docs.labelbox.com/en/configure-editor/queue-system#reservation-system

        Args:
            data_rows: data row identifiers object to update priorities.
                DataRowIdentifier objects are lists of ids or global keys. A DataIdentifier object can be a UniqueIds or GlobalKeys class.
            priority (int): Priority for the new override. See above for more information.

        Returns:
            bool, indicates if the operation was a success.
        """

        if not isinstance(data_rows, get_args(DataRowIdentifiers)):
            raise TypeError("data_rows must be a DataRowIdentifiers object")

        method = "createQueuePriorityUpdateTask"
        priority_param = "priority"
        project_param = "projectId"
        data_rows_param = "dataRowIdentifiers"
        query_str = """mutation %sPyApi(
              $%s: Int!
              $%s: ID!
              $%s: QueuePriorityUpdateDataRowIdentifiersInput
            ) {
              project(where: { id: $%s }) {
                %s(
                  data: { priority: $%s, dataRowIdentifiers: $%s }
                ) {
                  taskId
                }
              }
            }
        """ % (
            method,
            priority_param,
            project_param,
            data_rows_param,
            project_param,
            method,
            priority_param,
            data_rows_param,
        )
        res = self.client.execute(
            query_str,
            {
                priority_param: priority,
                project_param: self.uid,
                data_rows_param: {
                    "ids": [id for id in data_rows],
                    "idType": data_rows.id_type,
                },
            },
        )["project"][method]

        task_id = res["taskId"]

        task = self._wait_for_task(task_id)
        if task.status != "COMPLETE":
            raise LabelboxError(
                "Priority was not updated successfully: "
                + json.dumps(task.errors)
            )
        return True

    def extend_reservations(self, queue_type) -> int:
        """Extends all the current reservations for the current user on the given
        queue type.
        Args:
            queue_type (str): Either "LabelingQueue" or "ReviewQueue"
        Returns:
            int, the number of reservations that were extended.
        """
        if queue_type not in ("LabelingQueue", "ReviewQueue"):
            raise InvalidQueryError("Unsupported queue type: %s" % queue_type)

        id_param = "projectId"
        query_str = """mutation ExtendReservationsPyApi($%s: ID!){
            extendReservations(projectId:$%s queueType:%s)}""" % (
            id_param,
            id_param,
            queue_type,
        )
        res = self.client.execute(query_str, {id_param: self.uid})
        return res["extendReservations"]

    def enable_model_assisted_labeling(self, toggle: bool = True) -> bool:
        """Turns model assisted labeling either on or off based on input

        Args:
            toggle (bool): True or False boolean
        Returns:
            True if toggled on or False if toggled off
        """
        project_param = "project_id"
        show_param = "show"

        query_str = """mutation toggle_model_assisted_labelingPyApi($%s: ID!, $%s: Boolean!) {
            project(where: {id: $%s }) {
                showPredictionsToLabelers(show: $%s) {
                    id, showingPredictionsToLabelers
                }
            }
        }""" % (project_param, show_param, project_param, show_param)

        params = {project_param: self.uid, show_param: toggle}

        res = self.client.execute(query_str, params)
        return res["project"]["showPredictionsToLabelers"][
            "showingPredictionsToLabelers"
        ]

    def batches(self) -> PaginatedCollection:
        """Fetch all batches that belong to this project

        Returns:
            A `PaginatedCollection` of `Batch`es
        """
        id_param = "projectId"
        query_str = """query GetProjectBatchesPyApi($from: String, $first: PageSize, $%s: ID!) {
            project(where: {id: $%s}) {id
            batches(after: $from, first: $first) { nodes { %s } pageInfo { endCursor }}}}
        """ % (id_param, id_param, query.results_query_part(Entity.Batch))
        return PaginatedCollection(
            self.client,
            query_str,
            {id_param: self.uid},
            ["project", "batches", "nodes"],
            lambda client, res: Entity.Batch(client, self.uid, res),
            cursor_path=["project", "batches", "pageInfo", "endCursor"],
            experimental=True,
        )

    def task_queues(self) -> List[TaskQueue]:
        """Fetch all task queues that belong to this project

        Returns:
            A `List` of `TaskQueue`s
        """
        query_str = """query GetProjectTaskQueuesPyApi($projectId: ID!) {
              project(where: {id: $projectId}) {
                taskQueues {
                  %s
                }
              }
            }
        """ % (query.results_query_part(Entity.TaskQueue))

        task_queue_values = self.client.execute(
            query_str, {"projectId": self.uid}, timeout=180.0, experimental=True
        )["project"]["taskQueues"]

        return [
            Entity.TaskQueue(self.client, field_values)
            for field_values in task_queue_values
        ]

    def move_data_rows_to_task_queue(
        self,
        data_row_ids: DataRowIdentifiers,
        task_queue_id: Optional[str] = None,
    ):
        """

        Moves data rows to the specified task queue.

        Args:
            data_row_ids: a list of data row ids to be moved. This should be a DataRowIdentifiers object
                DataRowIdentifier objects are lists of ids or global keys. A DataIdentifier object can be a UniqueIds or GlobalKeys class.
            task_queue_id: the task queue id to be moved to, or None to specify the "Done" queue. Defaults to None.

        Returns:
            None if successful, or a raised error on failure

        """

        if not isinstance(data_row_ids, get_args(DataRowIdentifiers)):
            raise TypeError("data_rows must be a DataRowIdentifiers object")

        method = "createBulkAddRowsToQueueTask"
        query_str = (
            """mutation AddDataRowsToTaskQueueAsyncPyApi(
          $projectId: ID!
          $queueId: ID
          $dataRowIdentifiers: AddRowsToTaskQueueViaDataRowIdentifiersInput!
        ) {
          project(where: { id: $projectId }) {
            %s(
              data: { queueId: $queueId, dataRowIdentifiers: $dataRowIdentifiers }
            ) {
              taskId
            }
          }
        }
        """
            % method
        )

        task_id = self.client.execute(
            query_str,
            {
                "projectId": self.uid,
                "queueId": task_queue_id,
                "dataRowIdentifiers": {
                    "ids": [id for id in data_row_ids],
                    "idType": data_row_ids.id_type,
                },
            },
            timeout=180.0,
            experimental=True,
        )["project"][method]["taskId"]

        task = self._wait_for_task(task_id)
        if task.status != "COMPLETE":
            raise LabelboxError(
                "Data rows were not moved successfully: "
                + json.dumps(task.errors)
            )

    def _wait_for_task(self, task_id: str) -> Task:
        task = Task.get_task(self.client, task_id)
        task.wait_till_done()

        return task

    def _wait_until_data_rows_are_processed(
        self,
        data_row_ids: Optional[List[str]] = None,
        global_keys: Optional[List[str]] = None,
        wait_processing_max_seconds: int = _wait_processing_max_seconds,
        sleep_interval=30,
    ):
        """Wait until all the specified data rows are processed"""
        start_time = datetime.now()

        max_data_rows_per_poll = 100_000
        if data_row_ids is not None:
            for i in range(0, len(data_row_ids), max_data_rows_per_poll):
                chunk = data_row_ids[i : i + max_data_rows_per_poll]
                self._poll_data_row_processing_status(
                    chunk,
                    [],
                    start_time,
                    wait_processing_max_seconds,
                    sleep_interval,
                )

        if global_keys is not None:
            for i in range(0, len(global_keys), max_data_rows_per_poll):
                chunk = global_keys[i : i + max_data_rows_per_poll]
                self._poll_data_row_processing_status(
                    [],
                    chunk,
                    start_time,
                    wait_processing_max_seconds,
                    sleep_interval,
                )

    def _poll_data_row_processing_status(
        self,
        data_row_ids: List[str],
        global_keys: List[str],
        start_time: datetime,
        wait_processing_max_seconds: int = _wait_processing_max_seconds,
        sleep_interval=30,
    ):
        while True:
            if (
                datetime.now() - start_time
            ).total_seconds() >= wait_processing_max_seconds:
                raise ProcessingWaitTimeout(
                    """Maximum wait time exceeded while waiting for data rows to be processed.
                    Try creating a batch a bit later"""
                )

            all_good = self.__check_data_rows_have_been_processed(
                data_row_ids, global_keys
            )
            if all_good:
                return

            logger.debug(
                "Some of the data rows are still being processed, waiting..."
            )
            time.sleep(sleep_interval)

    def __check_data_rows_have_been_processed(
        self,
        data_row_ids: Optional[List[str]] = None,
        global_keys: Optional[List[str]] = None,
    ):
        if data_row_ids is not None and len(data_row_ids) > 0:
            param_name = "dataRowIds"
            params = {param_name: data_row_ids}
        else:
            param_name = "globalKeys"
            global_keys = global_keys if global_keys is not None else []
            params = {param_name: global_keys}

        query_str = """query CheckAllDataRowsHaveBeenProcessedPyApi($%s: [ID!]) {
            queryAllDataRowsHaveBeenProcessed(%s:$%s) {
                allDataRowsHaveBeenProcessed
           }
        }""" % (param_name, param_name, param_name)

        response = self.client.execute(query_str, params)
        return response["queryAllDataRowsHaveBeenProcessed"][
            "allDataRowsHaveBeenProcessed"
        ]

    def get_overview(
        self, details=False
    ) -> Union[ProjectOverview, ProjectOverviewDetailed]:
        """Return the overview of a project.

        This method returns the number of data rows per task queue and issues of a project,
        which is equivalent to the Overview tab of a project.

        Args:
            details (bool, optional): Whether to include detailed queue information for review and rework queues.
                Defaults to False.

        Returns:
            Union[ProjectOverview, ProjectOverviewDetailed]: An object representing the project overview.
                If `details` is False, returns a `ProjectOverview` object.
                If `details` is True, returns a `ProjectOverviewDetailed` object.

        Raises:
            Exception: If there is an error executing the query.

        """
        query = """query ProjectGetOverviewPyApi($projectId: ID!) {
            project(where: { id: $projectId }) {      
            workstreamStateCounts {
                state
                count
            }
            taskQueues {
                queueType
                name
                dataRowCount
            }
            issues {
                totalCount
            }
            completedDataRowCount
            }
        }
        """

        # Must use experimental to access "issues"
        result = self.client.execute(
            query, {"projectId": self.uid}, experimental=True
        )["project"]

        # Reformat category names
        overview = {
            utils.snake_case(st["state"]): st["count"]
            for st in result.get("workstreamStateCounts")
            if st["state"] != "NotInTaskQueue"
        }

        overview["issues"] = result.get("issues", {}).get("totalCount")

        # Rename categories
        overview["to_label"] = overview.pop("unlabeled")
        overview["total_data_rows"] = overview.pop("all")

        if not details:
            return ProjectOverview(**overview)
        else:
            # Build dictionary for queue details for review and rework queues
            for category in ["rework", "review"]:
                queues = [
                    {tq["name"]: tq.get("dataRowCount")}
                    for tq in result.get("taskQueues")
                    if tq.get("queueType") == f"MANUAL_{category.upper()}_QUEUE"
                ]

                overview[f"in_{category}"] = {
                    "data": queues,
                    "total": overview[f"in_{category}"],
                }

            return ProjectOverviewDetailed(**overview)

    def clone(self) -> "Project":
        """
        Clones the current project.

        Returns:
            Project: The cloned project.
        """
        mutation = """
            mutation CloneProjectPyApi($projectId: ID!) {
                cloneProject(data: { projectId: $projectId }) {
                    id
                }
            }
        """
        result = self.client.execute(mutation, {"projectId": self.uid})
        return self.client.get_project(result["cloneProject"]["id"])

    def get_labeling_service(self) -> LabelingService:
        """Get the labeling service for this project.

        Will automatically create a labeling service if one does not exist.

        Returns:
            LabelingService: The labeling service for this project.
        """
        return LabelingService.getOrCreate(self.client, self.uid)

    def get_labeling_service_status(self) -> LabelingServiceStatus:
        """Get the labeling service status for this project.

        Raises:
            ResourceNotFoundError if the project does not have a labeling service.

        Returns:
            LabelingServiceStatus: The labeling service status for this project.
        """
        return self.get_labeling_service().status

    def get_labeling_service_dashboard(self) -> LabelingServiceDashboard:
        """Get the labeling service for this project.

        Returns:
            LabelingServiceDashboard: The labeling service for this project.

        Attributes of the dashboard include:
            id (str): The project id.
            name (str): The project name.
            created_at, updated_at (datetime): The creation and last update times of the labeling service. None if labeling service is not requested.
            created_by_id (str): The user id of the creator of the labeling service. None if labeling service is not requested.
            status (LabelingServiceStatus): The status of the labeling service. Returns LabelingServiceStatus.Missing if labeling service is not requested.
            data_rows_count (int): The number of data rows in the project. 0 if labeling service is not requested.
            data_rows_in_review_count (int): The number of data rows in review queue. 0 if labeling service is not requested.
            data_rows_in_rework_count (int): The number of data rows in rework. 0 if labeling service is not requested.
            data_rows_done_count (int): The number of data rows in done queue. 0 if labeling service is not requested.
            tags (List[str]): Project tags.
            tasks_completed (int): The number of tasks completed, the same as data_rows_done_count. 0 if labeling service is not requested.
            tasks_remaining (int): The number of tasks remaining, the same as data_rows_count - data_rows_done_count. 0 if labeling service is not requested.
            service_type (str): Descriptive type for labeling service.

            NOTE can call dict() to get all attributes as dictionary.

        """
        return LabelingServiceDashboard.get(self.client, self.uid)


class ProjectMember(DbObject):
    user = Relationship.ToOne("User", cache=True)
    role = Relationship.ToOne("Role", cache=True)
    access_from = Field.String("access_from")


class LabelingParameterOverride(DbObject):
    """Customizes the order of assets in the label queue.

    Attributes:
        priority (int): A prioritization score.
        number_of_labels (int): Number of times an asset should be labeled.
    """

    priority = Field.Int("priority")
    number_of_labels = Field.Int("number_of_labels")

    data_row = Relationship.ToOne("DataRow", cache=True)


LabelerPerformance = namedtuple(
    "LabelerPerformance",
    "user count seconds_per_label, total_time_labeling "
    "consensus average_benchmark_agreement last_activity_time",
)
LabelerPerformance.__doc__ = (
    "Named tuple containing info about a labeler's performance."
)



from unittest.mock import patch

import pytest

from labelbox import MediaType


@pytest.mark.parametrize(
    "prompt_response_ontology, prompt_response_generation_project_with_new_dataset",
    [
        (MediaType.LLMPromptCreation, MediaType.LLMPromptCreation),
        (
            MediaType.LLMPromptResponseCreation,
            MediaType.LLMPromptResponseCreation,
        ),
    ],
    indirect=True,
)
def test_prompt_response_generation_ontology_project(
    client,
    prompt_response_ontology,
    prompt_response_generation_project_with_new_dataset,
    response_data_row,
    rand_gen,
):
    ontology = prompt_response_ontology

    assert ontology
    assert ontology.name

    for classification in ontology.classifications():
        assert classification.schema_id
        assert classification.feature_schema_id

    project = prompt_response_generation_project_with_new_dataset

    project.connect_ontology(ontology)

    assert project.labeling_frontend().name == "Editor"
    assert project.ontology().name == ontology.name

    with pytest.raises(
        ValueError,
        match="Cannot create batches for auto data generation projects",
    ):
        project.create_batch(
            rand_gen(str),
            [response_data_row.uid],  # sample of data row objects
        )

    with pytest.raises(
        ValueError,
        match="Cannot create batches for auto data generation projects",
    ):
        with patch(
            "labelbox.schema.project.MAX_SYNC_BATCH_ROW_COUNT", new=0
        ):  # force to async
            project.create_batch(
                rand_gen(str),
                [response_data_row.uid],  # sample of data row objects
            )


@pytest.mark.parametrize(
    "prompt_response_ontology, prompt_response_generation_project_with_dataset_id",
    [
        (MediaType.LLMPromptCreation, MediaType.LLMPromptCreation),
        (
            MediaType.LLMPromptResponseCreation,
            MediaType.LLMPromptResponseCreation,
        ),
    ],
    indirect=True,
)
def test_prompt_response_generation_ontology_project_with_existing_dataset(
    prompt_response_ontology, prompt_response_generation_project_with_dataset_id
):
    ontology = prompt_response_ontology

    project = prompt_response_generation_project_with_dataset_id
    assert project
    project.connect_ontology(ontology)

    assert project.labeling_frontend().name == "Editor"
    assert project.ontology().name == ontology.name


@pytest.fixture
def classification_json():
    classifications = [
        {
            "featureSchemaId": None,
            "kind": "Prompt",
            "minCharacters": 2,
            "maxCharacters": 10,
            "name": "prompt text",
            "instructions": "prompt text",
            "required": True,
            "schemaNodeId": None,
            "scope": "global",
            "type": "prompt",
            "options": [],
        },
        {
            "featureSchemaId": None,
            "kind": "ResponseCheckboxQuestion",
            "name": "response checklist",
            "instructions": "response checklist",
            "options": [
                {
                    "featureSchemaId": None,
                    "kind": "ResponseCheckboxOption",
                    "label": "response checklist option",
                    "schemaNodeId": None,
                    "position": 0,
                    "value": "option_1",
                }
            ],
            "required": True,
            "schemaNodeId": None,
            "scope": "global",
            "type": "response-checklist",
        },
        {
            "featureSchemaId": None,
            "kind": "ResponseText",
            "maxCharacters": 10,
            "minCharacters": 1,
            "name": "response text",
            "instructions": "response text",
            "required": True,
            "schemaNodeId": None,
            "scope": "global",
            "type": "response-text",
            "options": [],
        },
    ]

    return classifications


@pytest.fixture
def features_from_json(client, classification_json):
    classifications = classification_json
    features = {client.create_feature_schema(t) for t in classifications if t}

    yield features

    for f in features:
        client.delete_unused_feature_schema(f.uid)


@pytest.fixture
def ontology_from_feature_ids(client, features_from_json):
    feature_ids = {f.uid for f in features_from_json}
    ontology = client.create_ontology_from_feature_schemas(
        name="test-prompt_response_creation{rand_gen(str)}",
        feature_schema_ids=feature_ids,
        media_type=MediaType.LLMPromptResponseCreation,
    )

    yield ontology

    client.delete_unused_ontology(ontology.uid)


def test_ontology_create_feature_schema(
    ontology_from_feature_ids, features_from_json, classification_json
):
    created_ontology = ontology_from_feature_ids
    feature_schema_ids = {f.uid for f in features_from_json}
    classifications_normalized = created_ontology.normalized["classifications"]
    classifications = classification_json

    for classification in classifications:
        generated_tool = next(
            c
            for c in classifications_normalized
            if c["name"] == classification["name"]
        )
        assert generated_tool["schemaNodeId"] is not None
        assert generated_tool["featureSchemaId"] in feature_schema_ids
        assert generated_tool["type"] == classification["type"]
        assert generated_tool["name"] == classification["name"]
        assert generated_tool["required"] == classification["required"]



import os
import pytest


@pytest.mark.skipif(
    not os.environ.get("LABELBOX_TEST_ENVIRON") == "ephemeral",
    reason="This test only runs in EPHEMERAL environment",
)
def test_org_and_user_setup(client, ephmeral_client):
    assert type(client) is ephmeral_client
    assert client.admin_client
    assert client.api_key != client.admin_client.api_key

    organization = client.get_organization()
    assert organization
    user = client.get_user()
    assert user


@pytest.mark.skipif(
    os.environ.get("LABELBOX_TEST_ENVIRON") == "ephemeral",
    reason="This test does not run in EPHEMERAL environment",
)
def test_integration_client(client, integration_client):
    assert type(client) is integration_client



import logging
from enum import Enum
from typing import Iterable

from labelbox.orm import query
from labelbox.orm.db_object import DbObject, Updateable
from labelbox.orm.model import Entity, Field, Relationship

logger = logging.getLogger(__name__)


class Webhook(DbObject, Updateable):
    """Represents a server-side rule for sending notifications to a web-server
    whenever one of several predefined actions happens within a context of
    a Project or an Organization.

    Attributes:
        updated_at (datetime)
        created_at (datetime)
        url (str)
        topics (str): LABEL_CREATED, LABEL_UPDATED, LABEL_DELETED
        status (str): ACTIVE, INACTIVE, REVOKED

    """

    class Status(Enum):
        ACTIVE = "ACTIVE"
        INACTIVE = "INACTIVE"
        REVOKED = "REVOKED"

    class Topic(Enum):
        LABEL_CREATED = "LABEL_CREATED"
        LABEL_UPDATED = "LABEL_UPDATED"
        LABEL_DELETED = "LABEL_DELETED"
        WORKFLOW_ACTION = "WORKFLOW_ACTION"

    # For backwards compatibility
    for topic in Status:
        vars()[topic.name] = topic.value

    for status in Topic:
        vars()[status.name] = status.value

    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    url = Field.String("url")
    topics = Field.String("topics")
    status = Field.String("status")

    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization")
    project = Relationship.ToOne("Project")

    @staticmethod
    def create(client, topics, url, secret, project) -> "Webhook":
        """Creates a Webhook.

        Args:
            client (Client): The Labelbox client used to connect
                to the server.
            topics (list of str): A list of topics this Webhook should
                get notifications for. Must be one of Webhook.Topic
            url (str): The URL to which notifications should be sent
                by the Labelbox server.
            secret (str): A secret key used for signing notifications.
            project (Project or None): The project for which notifications
                should be sent. If None notifications are sent for all
                events in your organization.
        Returns:
            A newly created Webhook.

        Raises:
            ValueError: If the topic is not one of Topic or status is not one of Status

        Information on configuring your server can be found here (this is where the url points to and the secret is set).
                        https://docs.labelbox.com/reference/webhook

        """
        if not secret:
            raise ValueError("Secret must be a non-empty string.")
        if not topics:
            raise ValueError("Topics must be a non-empty list.")
        if not url:
            raise ValueError("URL must be a non-empty string.")
        Webhook.validate_topics(topics)

        project_str = (
            "" if project is None else ('project:{id:"%s"},' % project.uid)
        )

        query_str = """mutation CreateWebhookPyApi {
            createWebhook(data:{%s topics:{set:[%s]}, url:"%s", secret:"%s" }){%s}
        } """ % (
            project_str,
            " ".join(topics),
            url,
            secret,
            query.results_query_part(Entity.Webhook),
        )

        return Webhook(client, client.execute(query_str)["createWebhook"])

    @staticmethod
    def validate_topics(topics) -> None:
        if isinstance(topics, str) or not isinstance(topics, Iterable):
            raise TypeError(
                f"Topics must be List[Webhook.Topic]. Found `{topics}`"
            )

        for topic in topics:
            Webhook.validate_value(topic, Webhook.Topic)

    @staticmethod
    def validate_value(value, enum) -> None:
        supported_values = {item.value for item in enum}
        if value not in supported_values:
            raise ValueError(
                f"Value `{value}` does not exist in supported values. Expected one of {supported_values}"
            )

    def delete(self) -> None:
        """
        Deletes the webhook
        """
        self.update(status=self.Status.INACTIVE.value)

    def update(self, topics=None, url=None, status=None):
        """Updates the Webhook.

        Args:
            topics (Optional[List[Topic]]): The new topics.
            url  Optional[str): The new URL value.
            status (Optional[Status]): The new status.
                If an argument is set to None then no updates will be made to that field.

        """

        # Webhook has a custom `update` function due to custom types
        # in `status` and `topics` fields.

        if topics is not None:
            self.validate_topics(topics)

        if status is not None:
            self.validate_value(status, self.Status)

        topics_str = (
            "" if topics is None else "topics: {set: [%s]}" % " ".join(topics)
        )
        url_str = "" if url is None else 'url: "%s"' % url
        status_str = "" if status is None else "status: %s" % status

        query_str = """mutation UpdateWebhookPyApi {
            updateWebhook(where: {id: "%s"} data:{%s}){%s}} """ % (
            self.uid,
            ", ".join(filter(None, (topics_str, url_str, status_str))),
            query.results_query_part(Entity.Webhook),
        )

        self._set_field_values(self.client.execute(query_str)["updateWebhook"])



import json
import os
import tempfile
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
import warnings

import requests
from pydantic import BaseModel

from labelbox.schema.task import Task
from labelbox.utils import _CamelCaseMixin

if TYPE_CHECKING:
    from labelbox import Client

OutputT = TypeVar("OutputT")


class StreamType(Enum):
    """The type of the stream."""

    RESULT = "RESULT"
    ERRORS = "ERRORS"


class Range(_CamelCaseMixin, BaseModel):  # pylint: disable=too-few-public-methods
    """Represents a range."""

    start: int
    end: int


class _MetadataHeader(_CamelCaseMixin, BaseModel):  # pylint: disable=too-few-public-methods
    total_size: int
    total_lines: int


class _MetadataFileInfo(_CamelCaseMixin, BaseModel):  # pylint: disable=too-few-public-methods
    offsets: Range
    lines: Range
    file: str


@dataclass
class _TaskContext:
    client: "Client"
    task_id: str
    stream_type: StreamType
    metadata_header: _MetadataHeader


class Converter(ABC, Generic[OutputT]):
    """Abstract class for transforming data."""

    @dataclass
    class ConverterInputArgs:
        """Input for the converter."""

        ctx: _TaskContext
        file_info: _MetadataFileInfo
        raw_data: str

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    @abstractmethod
    def convert(self, input_args: ConverterInputArgs) -> Iterator[OutputT]:
        """Converts the data.
        Returns an iterator that yields the converted data.

        Args:
            current_offset: The global offset indicating the position of the data within the
                            exported files. It represents a cumulative offset in characters
                            across multiple files.
            raw_data: The raw data to convert.

        Yields:
            Iterator[OutputT]: The converted data.
        """


class FileRetrieverStrategy(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for retrieving files."""

    def __init__(self, ctx: _TaskContext) -> None:
        super().__init__()
        self._ctx = ctx

    @abstractmethod
    def get_next_chunk(self) -> Optional[Tuple[_MetadataFileInfo, str]]:
        """Retrieves the file."""

    def _get_file_content(
        self, query: str, variables: dict, result_field_name: str
    ) -> Tuple[_MetadataFileInfo, str]:
        """Runs the query."""
        res = self._ctx.client.execute(query, variables, error_log_key="errors")
        res = res["task"][result_field_name]
        file_info = _MetadataFileInfo(**res) if res else None
        if not file_info:
            raise ValueError(
                f"Task {self._ctx.task_id} does not have a metadata file for the "
                f"{self._ctx.stream_type.value} stream"
            )
        response = requests.get(file_info.file, timeout=30)
        response.raise_for_status()
        response.encoding = "utf-8"
        assert (
            len(response.content)
            == file_info.offsets.end - file_info.offsets.start + 1
        ), (
            f"expected {file_info.offsets.end - file_info.offsets.start + 1} bytes, "
            f"got {len(response.content)} bytes"
        )
        return file_info, response.text


class _Reader(ABC):  # pylint: disable=too-few-public-methods
    """Abstract class for reading data from a source."""

    @abstractmethod
    def set_retrieval_strategy(self, strategy: FileRetrieverStrategy) -> None:
        """Sets the retrieval strategy."""

    @abstractmethod
    def read(self) -> Iterator[Tuple[_MetadataFileInfo, str]]:
        """Reads data from the source."""


class _BufferedFileRetrieverByOffset(FileRetrieverStrategy):  # pylint: disable=too-few-public-methods
    """Retrieves files by offset."""

    def __init__(
        self,
        ctx: _TaskContext,
        offset: int,
    ) -> None:
        super().__init__(ctx)
        self._current_offset = offset
        self._current_line = 0
        if self._current_offset >= self._ctx.metadata_header.total_size:
            raise ValueError(
                f"offset is out of range, max offset is {self._ctx.metadata_header.total_size - 1}"
            )

    def get_next_chunk(self) -> Optional[Tuple[_MetadataFileInfo, str]]:
        if self._current_offset >= self._ctx.metadata_header.total_size:
            return None
        query = (
            f"query GetExportFileFromOffsetPyApi"
            f"($where: WhereUniqueIdInput, $streamType: TaskStreamType!, $offset: UInt64!)"
            f"{{task(where: $where)"
            f"{{{'exportFileFromOffset'}(streamType: $streamType, offset: $offset)"
            f"{{offsets {{start end}} lines {{start end}} file}}"
            f"}}}}"
        )
        variables = {
            "where": {"id": self._ctx.task_id},
            "streamType": self._ctx.stream_type.value,
            "offset": str(self._current_offset),
        }
        file_info, file_content = self._get_file_content(
            query, variables, "exportFileFromOffset"
        )
        file_info.offsets.start = self._current_offset
        file_info.lines.start = self._current_line
        self._current_offset = file_info.offsets.end + 1
        self._current_line = file_info.lines.end + 1
        return file_info, file_content


class BufferedStream(Generic[OutputT]):
    """Streams data from a Reader."""

    def __init__(
        self,
        ctx: _TaskContext,
    ):
        self._ctx = ctx
        self._reader = _BufferedGCSFileReader()
        self._converter = _BufferedJsonConverter()
        self._reader.set_retrieval_strategy(
            _BufferedFileRetrieverByOffset(self._ctx, 0)
        )

    def __iter__(self):
        yield from self._fetch()

    def _fetch(
        self,
    ) -> Iterator[OutputT]:
        """Fetches the result data.
        Returns an iterator that yields the offset and the data.
        """
        if self._ctx.metadata_header.total_size is None:
            return

        stream = self._reader.read()
        with self._converter as converter:
            for file_info, raw_data in stream:
                for output in converter.convert(
                    Converter.ConverterInputArgs(self._ctx, file_info, raw_data)
                ):
                    yield output

    def start(
        self, stream_handler: Optional[Callable[[OutputT], None]] = None
    ) -> None:
        """Starts streaming the result data.
        Calls the stream_handler for each result.
        """
        # this calls the __iter__ method, which in turn calls the _fetch method
        for output in self:
            if stream_handler:
                stream_handler(output)


@dataclass
class BufferedJsonConverterOutput:
    """Output with the JSON object"""

    json: Any


class _BufferedJsonConverter(Converter[BufferedJsonConverterOutput]):
    """Converts JSON data in a buffered manner"""

    def convert(
        self, input_args: Converter.ConverterInputArgs
    ) -> Iterator[BufferedJsonConverterOutput]:
        yield BufferedJsonConverterOutput(json=json.loads(input_args.raw_data))


class _BufferedGCSFileReader(_Reader):
    """Reads data from multiple GCS files and buffer them to disk"""

    def __init__(self):
        super().__init__()
        self._retrieval_strategy = None

    def set_retrieval_strategy(self, strategy: FileRetrieverStrategy) -> None:
        """Sets the retrieval strategy."""
        self._retrieval_strategy = strategy

    def read(self) -> Iterator[Tuple[_MetadataFileInfo, str]]:
        if not self._retrieval_strategy:
            raise ValueError("retrieval strategy not set")
        # create a buffer
        with tempfile.NamedTemporaryFile(mode="w+", delete=False) as temp_file:
            result = self._retrieval_strategy.get_next_chunk()
            while result:
                _, raw_data = result
                # there is something wrong with the way the offsets are being calculated
                # so just write all of the chunks as is too the file, with pointer initially
                # pointed to the start of the file (like what is in GCS) and do not
                # rely on offsets for file location
                # temp_file.seek(file_info.offsets.start)
                temp_file.write(raw_data)
                result = self._retrieval_strategy.get_next_chunk()
        # read buffer
        with open(temp_file.name, "r") as temp_file_reopened:
            for idx, line in enumerate(temp_file_reopened):
                yield (
                    _MetadataFileInfo(
                        offsets=Range(start=0, end=len(line) - 1),
                        lines=Range(start=idx, end=idx + 1),
                        file=temp_file.name,
                    ),
                    line,
                )
        # manually delete buffer
        os.unlink(temp_file.name)


class ExportTask:
    """
    An adapter class for working with task objects, providing extended functionality
    and convenient access to task-related information.

    This class wraps a `Task` object, allowing you to interact with tasks of this type.
    It offers methods to retrieve task results, errors, and metadata, as well as properties
    for accessing task details such as UID, status, and creation time.
    """

    class ExportTaskException(Exception):
        """Raised when the task is not ready yet."""

    def __init__(self, task: Task, is_export_v2: bool = False) -> None:
        self._is_export_v2 = is_export_v2
        self._task = task

    def __repr__(self):
        return (
            f"<ExportTask ID: {self.uid}>"
            if getattr(self, "uid", None)
            else "<ExportTask>"
        )

    def __str__(self):
        properties_to_include = [
            "completion_percentage",
            "created_at",
            "metadata",
            "name",
            "result",
            "result_url",
            "errors",
            "errors_url",
            "status",
            "type",
            "uid",
            "updated_at",
        ]
        props = {prop: getattr(self, prop) for prop in properties_to_include}
        return f"<ExportTask {json.dumps(props, indent=4, default=str)}>"

    def __eq__(self, other):
        return self._task.__eq__(other)

    def __hash__(self):
        return self._task.__hash__()

    @property
    def uid(self):
        """Returns the uid of the task."""
        return self._task.uid

    @property
    def deleted(self):
        """Returns whether the task is deleted."""
        return self._task.deleted

    @property
    def updated_at(self):
        """Returns the last time the task was updated."""
        return self._task.updated_at

    @property
    def created_at(self):
        """Returns the time the task was created."""
        return self._task.created_at

    @property
    def name(self):
        """Returns the name of the task."""
        return self._task.name

    @property
    def status(self):
        """Returns the status of the task."""
        return self._task.status

    @property
    def metadata(self):
        """Returns the metadata of the task."""
        return self._task.metadata

    @property
    def result_url(self):
        """Returns the result URL of the task."""
        if not self._is_export_v2:
            raise ExportTask.ExportTaskException(
                "This property is only available for export_v2 tasks due to compatibility reasons, please use streamable errors instead"
            )
        base_url = self._task.client.rest_endpoint
        return (
            base_url
            + "/export-results/"
            + self._task.uid
            + "/"
            + self._task.client.get_organization().uid
        )

    @property
    def errors_url(self):
        """Returns the errors URL of the task."""
        if not self._is_export_v2:
            raise ExportTask.ExportTaskException(
                "This property is only available for export_v2 tasks due to compatibility reasons, please use streamable errors instead"
            )
        if not self.has_errors():
            return None
        base_url = self._task.client.rest_endpoint
        return (
            base_url
            + "/export-errors/"
            + self._task.uid
            + "/"
            + self._task.client.get_organization().uid
        )

    @property
    def errors(self):
        """Returns the errors of the task."""
        if not self._is_export_v2:
            raise ExportTask.ExportTaskException(
                "This property is only available for export_v2 tasks due to compatibility reasons, please use streamable errors instead"
            )
        if self.status == "FAILED":
            raise ExportTask.ExportTaskException("Task failed")
        if self.status != "COMPLETE":
            raise ExportTask.ExportTaskException("Task is not ready yet")

        if not self.has_errors():
            return None

        data = []

        metadata_header = ExportTask._get_metadata_header(
            self._task.client, self._task.uid, StreamType.ERRORS
        )
        if metadata_header is None:
            return None
        BufferedStream(
            _TaskContext(
                self._task.client,
                self._task.uid,
                StreamType.ERRORS,
                metadata_header,
            ),
        ).start(stream_handler=lambda output: data.append(output.json))
        return data

    @property
    def result(self):
        """Returns the result of the task."""
        if self._is_export_v2:
            if self.status == "FAILED":
                raise ExportTask.ExportTaskException("Task failed")
            if self.status != "COMPLETE":
                raise ExportTask.ExportTaskException("Task is not ready yet")
            data = []

            metadata_header = ExportTask._get_metadata_header(
                self._task.client, self._task.uid, StreamType.RESULT
            )
            if metadata_header is None:
                return []
            BufferedStream(
                _TaskContext(
                    self._task.client,
                    self._task.uid,
                    StreamType.RESULT,
                    metadata_header,
                ),
            ).start(stream_handler=lambda output: data.append(output.json))
            return data
        return self._task.result_url

    @property
    def completion_percentage(self):
        """Returns the completion percentage of the task."""
        return self._task.completion_percentage

    @property
    def type(self):
        """Returns the type of the task."""
        return self._task.type

    @property
    def created_by(self):
        """Returns the user who created the task."""
        return self._task.created_by

    @property
    def organization(self):
        """Returns the organization of the task."""
        return self._task.organization

    def wait_until_done(self, timeout_seconds: int = 7200) -> None:
        warnings.warn(
            "The method wait_until_done for ExportTask is deprecated and will be removed in the next major release. Use the wait_till_done method instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.wait_till_done(timeout_seconds)

    def wait_till_done(self, timeout_seconds: int = 7200) -> None:
        """Waits until the task is done."""
        return self._task.wait_till_done(timeout_seconds)

    @staticmethod
    @lru_cache(maxsize=5)
    def _get_metadata_header(
        client, task_id: str, stream_type: StreamType
    ) -> Union[_MetadataHeader, None]:
        """Returns the total file size for a specific task."""
        query = (
            f"query GetExportMetadataHeaderPyApi"
            f"($where: WhereUniqueIdInput, $streamType: TaskStreamType!)"
            f"{{task(where: $where)"
            f"{{{'exportMetadataHeader'}(streamType: $streamType)"
            f"{{totalSize totalLines}}"
            f"}}}}"
        )
        variables = {"where": {"id": task_id}, "streamType": stream_type.value}
        res = client.execute(query, variables, error_log_key="errors")
        res = res["task"]["exportMetadataHeader"]
        return _MetadataHeader(**res) if res else None

    def get_total_file_size(self, stream_type: StreamType) -> Union[int, None]:
        """Returns the total file size for a specific task."""
        if self._task.status == "FAILED":
            raise ExportTask.ExportTaskException("Task failed")
        if self._task.status != "COMPLETE":
            raise ExportTask.ExportTaskException("Task is not ready yet")
        header = ExportTask._get_metadata_header(
            self._task.client, self._task.uid, stream_type
        )
        return header.total_size if header else None

    def get_total_lines(self, stream_type: StreamType) -> Union[int, None]:
        """Returns the total file size for a specific task."""
        if self._task.status == "FAILED":
            raise ExportTask.ExportTaskException("Task failed")
        if self._task.status != "COMPLETE":
            raise ExportTask.ExportTaskException("Task is not ready yet")
        header = ExportTask._get_metadata_header(
            self._task.client, self._task.uid, stream_type
        )
        return header.total_lines if header else None

    def has_result(self) -> bool:
        """Returns whether the task has a result."""
        total_size = self.get_total_file_size(StreamType.RESULT)
        return total_size is not None and total_size > 0

    def has_errors(self) -> bool:
        """Returns whether the task has errors."""
        total_size = self.get_total_file_size(StreamType.ERRORS)
        return total_size is not None and total_size > 0

    def get_buffered_stream(
        self,
        stream_type: StreamType = StreamType.RESULT,
    ) -> BufferedStream:
        """
        Returns the result of the task.

        Args:
            stream_type (StreamType, optional): The type of stream to retrieve. Defaults to StreamType.RESULT.

        Returns:
            Stream: The buffered stream object.

        Raises:
            ExportTask.ExportTaskException: If the task has failed or is not ready yet.
            ValueError: If the task does not have the specified stream type.
        """
        if self._task.status == "FAILED":
            raise ExportTask.ExportTaskException("Task failed")
        if self._task.status != "COMPLETE":
            raise ExportTask.ExportTaskException("Task is not ready yet")

        metadata_header = self._get_metadata_header(
            self._task.client, self._task.uid, stream_type
        )
        if metadata_header is None:
            raise ValueError(
                f"Task {self._task.uid} does not have a {stream_type.value} stream"
            )
        return BufferedStream(
            _TaskContext(
                self._task.client, self._task.uid, stream_type, metadata_header
            ),
        )

    @staticmethod
    def get_task(client, task_id):
        """Returns the task with the given id."""
        return ExportTask(Task.get_task(client, task_id))
